/**
 *  $Id: compile.c 225 2005-08-13 00:39:47Z ko1 $
 * Create : K.S. 04/01/01 03:42:15
 *
 * ruby node tree -> yarv code
 */


#include <ruby.h>
#include <node.h>
#include <env.h>

#include "yarvcore.h"
#include "compile.h"
#include "insns.inc"


#ifdef HAVE_STDARG_PROTOTYPES
#include <stdarg.h>
#define va_init_list(a,b) va_start(a,b)
#else
#include <varargs.h>
#define va_init_list(a,b) va_start(a)
#endif

/* types */

#define ISEQ_ELEMENT_NONE  INT2FIX(0x00)
#define ISEQ_ELEMENT_LABEL INT2FIX(0x01)
#define ISEQ_ELEMENT_INSN  INT2FIX(0x02)
#define ISEQ_ELEMENT_SEQ   INT2FIX(0x03)

typedef struct iseq_link_element{
  int type;
  struct iseq_link_element *next;
  struct iseq_link_element *prev;
} LINK_ELEMENT;

typedef struct iseq_link_anchor{
  LINK_ELEMENT anchor;
  LINK_ELEMENT *last;
} LINK_ANCHOR;

typedef struct iseq_label_data{
  struct iseq_link_element link;
  int label_no;
  int position;
  int sc_state;
  int set;
} LABEL;

typedef struct iseq_insn_data{
  struct iseq_link_element link;
  int insn_id;
  int line_no;
  int operand_size;
  int sc_state;
  VALUE *operands;
} INSN;


/* for debug */
#if CPDEBUG > 0
static long gl_node_level = 0;
static long gl_tmp = 0;
static void debug_list(LINK_ANCHOR *anchor);
#endif

void dump_disasm_list(LINK_ELEMENT* elem);
static int insn_data_length(INSN *insnobj);
static int insn_data_line_no(INSN *insnobj);
static int insn_ret_num(int insn);

static void ADD_ELEM(LINK_ANCHOR *anchor, LINK_ELEMENT* elem);

static INSN *new_insn_body(yarv_iseq_t *iseqobj, int line_no,
                                         int insn_id, int argc, ...);
static LABEL *new_label_body(yarv_iseq_t *iseqobj, int line);

static int iseq_compile_each(VALUE self, LINK_ANCHOR *anchor, NODE* n, int);
static int iseq_setup(VALUE self, LINK_ANCHOR *anchor);

static int iseq_optimize(yarv_iseq_t *iseqobj, LINK_ANCHOR *anchor);
static int iseq_insns_unification(yarv_iseq_t *iseqobj,  LINK_ANCHOR *anchor);
static int set_sequence_stackcaching(yarv_iseq_t *iseqobj, LINK_ANCHOR *anchor);
static int set_sequence(yarv_iseq_t *iseqobj,  LINK_ANCHOR *anchor);

static int set_exception_table(yarv_iseq_t *iseqobj);
static int set_localtbl(yarv_iseq_t *iseqobj, ID* tbl);
static int set_arguments(VALUE self, yarv_iseq_t *iseqobj, LINK_ANCHOR *anchor, NODE *node);
static NODE* set_block_local_tbl(VALUE self, yarv_iseq_t *iseqobj, NODE *node, LINK_ANCHOR *anchor);
static int set_exception_tbl(yarv_iseq_t *iseqobj);
static int set_optargs_table(yarv_iseq_t *iseqobj);

#include "optinsn.inc"
#include "optunifs.inc"

#define REWIND_DSIZE()  1
#define MREWIND_DSIZE() 2


VALUE iseq_compile(VALUE self, VALUE narg){
  DECL_ANCHOR(list_anchor);
  yarv_iseq_t *iseqobj;
  NODE *node = (NODE*)narg;


  debugs("[compile step 1 (traverse each node)]\n");
  
  GetISeqVal(self, iseqobj);

  iseqobj->node = node;
  
  if(iseqobj->type == ISEQ_TYPE_BLOCK){
    node = set_block_local_tbl(self, iseqobj, node, list_anchor);
  }
  /*
    compile_each return nested array.
    ex) [insnA, [insnB, ope1], "label", ...]
   */
  if(node && nd_type(node) == NODE_SCOPE){
    /* with node scope */
    NODE *sn_body = node->nd_next; /* sn: scope node */
    NODE *ndargs  = 0;

    if(iseqobj->type != ISEQ_TYPE_BLOCK){
      set_localtbl(iseqobj, ((NODE*)node)->nd_tbl);
    }

    if(sn_body){
      switch(nd_type(sn_body)){
      case NODE_BLOCK:
        if(nd_type(sn_body->nd_head) == NODE_ARGS){
          /* some method attribute process */
          ndargs = sn_body->nd_head;
          set_arguments(self, iseqobj, list_anchor, ndargs);
          
          /* with sn_body->nd_head */
          if(iseqobj->type == ISEQ_TYPE_METHOD){
            iseqobj->rewind_frame_size = iseqobj->local_size + MREWIND_DSIZE();
            COMPILE(list_anchor, "normal method", sn_body->nd_next);
          }
          else if(iseqobj->type == ISEQ_TYPE_CLASS){
            iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();
            COMPILE(list_anchor, "class/module", sn_body->nd_next);
          }
          else{
            rb_bug("must be class or method");
          }

        }
        else{
          /* normal block */

          if(iseqobj->type == ISEQ_TYPE_CLASS){
            iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();
            COMPILE(list_anchor, "class/module", sn_body);
          }
          else if(iseqobj->type == ISEQ_TYPE_BLOCK){
            iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();
            COMPILE(list_anchor, "normal block", sn_body);
          }
          else{
            rb_bug("must be class or block");
          }
        }
        break;
      case NODE_ARGS:
        /* empty method */
        /* some method attribute process */
        debugs("empty method\n");
        
        set_arguments(self, iseqobj, list_anchor, sn_body);
        ADD_INSN(list_anchor, nd_line(sn_body), putnil);

        iseqobj->rewind_frame_size = iseqobj->local_size + MREWIND_DSIZE();
        break;

      default:
        COMPILE(list_anchor, "other scope", sn_body);
        
        iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();
        break;
      }
    }
    else{
      /* sn_body == 0 */
      ADD_INSN(list_anchor, 0, putnil);
      
      iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();
    }
  }
  else{
    if(iseqobj->type == ISEQ_TYPE_BLOCK){
      VALUE tmp;
      LABEL *start = iseqobj->compile_data->start_label = NEW_LABEL(0);
      LABEL *end   = iseqobj->compile_data->end_label   = NEW_LABEL(0);
      
      iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();

      ADD_CATCH_ENTRY(CATCH_TYPE_REDO, start, end, 0, 0, start);
      ADD_CATCH_ENTRY(CATCH_TYPE_NEXT, start, end, 0, 0, end);

      ADD_LABEL(list_anchor, iseqobj->compile_data->start_label);
      COMPILE(list_anchor, "block body", node);
      ADD_LABEL(list_anchor, iseqobj->compile_data->end_label);
    }
    else if(iseqobj->type == ISEQ_TYPE_TOP){
      set_localtbl(iseqobj, ruby_scope->local_tbl);
      iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();
      COMPILE(list_anchor, "top level node", node);
    }
    else if(iseqobj->type == ISEQ_TYPE_RESCUE){
      set_exception_tbl(iseqobj);
      iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();
      COMPILE(list_anchor, "rescue", node);
    }
    else if(iseqobj->type == ISEQ_TYPE_ENSURE){
      set_exception_tbl(iseqobj);
      iseqobj->rewind_frame_size = iseqobj->local_size + REWIND_DSIZE();
      COMPILE_POPED(list_anchor, "ensure", node);
    }
    else if(node == 0){
      COMPILE(list_anchor, "nil", node);
    }
    else{
      rb_bug("unknown scope");
    }
  }

  GC_CHECK();

  if(iseqobj->type == ISEQ_TYPE_RESCUE ||
     iseqobj->type == ISEQ_TYPE_ENSURE){
    
    ADD_INSN2(list_anchor, 0, getdynamic, I2F(1), I2F(0));
    ADD_INSN1(list_anchor, 0, throw, I2F(0));
  }
  else{
    ADD_INSN(list_anchor, 0, end);
  }

  return iseq_setup(self, list_anchor);
}


VALUE th_eval(void *);
static int iseq_translate_direct_threaded_code(yarv_iseq_t *iseq){
#if OPT_DIRECT_THREADED_CODE || OPT_CALL_THREADED_CODE

#if OPT_DIRECT_THREADED_CODE
  void **table = (void **)th_eval(0);
#else
  extern void **insns_address_table();
  void **table = get_insns_address_table();
#endif
  int i;
  
  iseq->iseq_encoded = ALLOC_N(VALUE, iseq->size);
  MEMCPY(iseq->iseq_encoded, iseq->iseq, VALUE, iseq->size);

  for(i=0; i<iseq->size; /* */){
    int insn = iseq->iseq_encoded[i];
    int len  = insn_len(insn);
    iseq->iseq_encoded[i] = (VALUE)table[insn];
    i += len;
  }
#else
  iseq->iseq_encoded = iseq->iseq;
#endif
  return COMPILE_OK;
}

/*********************************************/
/* definition of data structure for compiler */
/*********************************************/


static void *compile_data_alloc(yarv_iseq_t *iseqobj, int size){
  void *ptr = 0;
  struct iseq_compile_data_storage *storage = iseqobj->compile_data->storage_current;

  if(storage->pos + size > storage->size){
    unsigned long alloc_size = storage->size * 2;
    
  retry:
    if(alloc_size < size){
      alloc_size *= 2;
      goto retry;
    }
    storage->next = (void *)ALLOC_N(char, alloc_size +
                                    sizeof(struct iseq_compile_data_storage));
    storage = iseqobj->compile_data->storage_current = storage->next;

    storage->next = 0;
    storage->pos  = 0;
    storage->size = alloc_size;
    storage->buff = (char *)(&storage->buff + 1);
  }
  
  ptr = (void *)&storage->buff[storage->pos];
  storage->pos += size;
  return ptr;
}

static INSN *compile_data_alloc_insn(yarv_iseq_t *iseqobj){
  return (INSN*)compile_data_alloc(iseqobj, sizeof(INSN));
}

static LABEL *compile_data_alloc_label(yarv_iseq_t *iseqobj){
  return (LABEL*)compile_data_alloc(iseqobj, sizeof(LABEL));
}


/*
 * To make Array to LinkedList, use link_anchor
 */


static void verify_list(char *info, LINK_ANCHOR *anchor){
#if CPDEBUG > 0
  int flag = 0;
  LINK_ELEMENT* list = anchor->anchor.next,
                   * plist= &anchor->anchor;
  while(list){
    if(plist != list->prev){
      flag += 1;
    }
    plist = list;
    list  = list->next;
  }
  if(anchor->last != plist){
    flag |= 0x70000;
  }
  
  if(flag != 0){
    rb_bug("list verify error: %08x (%s)", flag, info);
  }
#endif
}

/*
 * elem1, elem2 => elem1, elem2, elem
 */
static void ADD_ELEM(LINK_ANCHOR *anchor, LINK_ELEMENT* elem){
  elem->prev = anchor->last;
  anchor->last->next = elem;
  anchor->last = elem;
  verify_list("add", anchor);
}

/*******************************************/
#if 0
/*
 * elem1, elemX => elem1, elem2, elemX
 */
static void INSERT_ELEM_NEXT(LINK_ELEMENT* elem1,
                             LINK_ELEMENT* elem2){
  elem2->next = elem1->next;
  elem2->prev = elem1;
  elem1->next = elem2;
  if(elem2->next){
    elem2->next->prev = elem2;
  }
}

/*
 * elemX, elem1 => elemX, elem2, elem1
 */
static void INSERT_ELEM_PREV(LINK_ELEMENT* elem1,
                             LINK_ELEMENT* elem2){
  elem2->prev = elem1->prev;
  elem2->next = elem1;
  elem1->prev = elem2;
  if(elem2->prev){
    elem2->prev->next = elem2;
  }
}
#endif
/*******************************************/

/*
 * elemX, elem1, elemY => elemX, elem2, elemY
 */
static void REPLACE_ELEM(LINK_ELEMENT* elem1,
                         LINK_ELEMENT* elem2){
  elem2->prev = elem1->prev;
  elem2->next = elem1->next;
  if(elem1->prev){
    elem1->prev->next = elem2;
  }
  if(elem1->next){
    elem1->next->prev = elem2;
  }
}

static void REMOVE_ELEM(LINK_ELEMENT *elem){
  elem->prev->next = elem->next;
  if(elem->next){
    elem->next->prev = elem->prev;
  }
}

static LINK_ELEMENT *FIRST_ELEMENT(LINK_ANCHOR *anchor){
  return anchor->anchor.next;
}

static LINK_ELEMENT *LAST_ELEMENT(LINK_ANCHOR *anchor){
  return anchor->last;
}

static LINK_ELEMENT *POP_ELEMENT(LINK_ANCHOR *anchor){
  LINK_ELEMENT *elem = anchor->last;
  anchor->last = anchor->last->prev;
  anchor->last->next = 0;
  verify_list("pop", anchor);
  return elem;
}

static int LIST_SIZE(LINK_ANCHOR *anchor){
  LINK_ELEMENT *elem = anchor->anchor.next;
  int size = 0;
  while(elem){
    size += 1;
    elem = elem->next;
  }
  return size;
}

/*
 * anc1: e1, e2, e3
 * anc2: e4, e5
 *#=>
 * anc1: e1, e2, e3, e4, e5
 * anc2: e4, e5 (broken)
 */
static void APPEND_LIST(LINK_ANCHOR *anc1, LINK_ANCHOR *anc2){
  if(anc2->anchor.next){
    anc1->last->next = anc2->anchor.next;
    anc2->anchor.next->prev = anc1->last;
    anc1->last = anc2->last;
  }
  verify_list("append", anc1);
}

#if 0
/*
 * anc1: e1, e2, e3
 * anc2: e4, e5
 *#=>
 * anc1: e4, e5
 * anc2: e1, e2, e3
 */
static void SWAP_LIST(LINK_ANCHOR *anc1, LINK_ANCHOR *anc2){
  LINK_ANCHOR tmp = *anc2;
  *anc2 = *anc1;
  *anc1 = tmp;
  verify_list("swap1", anc1);
  verify_list("swap2", anc2);
}
#endif

static LINK_ANCHOR *REVERSE_LIST(LINK_ANCHOR *anc){
  LINK_ELEMENT *first, *last, *elem, *e;
  first = &anc->anchor;
  elem  = first->next;
  last  = anc->last;

  if(elem != 0){
    anc->anchor.next = last;
    anc->last        = elem;
  }
  else{
    /* null list */
    return anc;
  }
  while(elem){
    e = elem->next;
    elem->next = elem->prev;
    elem->prev = e;
    elem = e;
  }

  first->next = last;
  last->prev = first;
  anc->last->next = 0;
  
  verify_list("reverse", anc);
  return anc;
}

#if CPDEBUG > 0
static void debug_list(LINK_ANCHOR *anchor){
  LINK_ELEMENT* list = FIRST_ELEMENT(anchor);
  int i = 0;
  printf("----\n");
  printf("anch: %p, frst: %p, last: %p\n", &anchor->anchor, anchor->anchor.next, anchor->last);
  while(list){
    printf("curr: %p, next: %p, prev: %p, type: %d\n", list, list->next, list->prev, FIX2INT(list->type));
    list = list->next;
  }
  printf("----\n");

  dump_disasm_list(anchor->anchor.next);
  verify_list("debug list", anchor);
}
#endif

static LABEL *new_label_body(yarv_iseq_t *iseqobj, int line){
  LABEL *labelobj = compile_data_alloc_label(iseqobj);
  static int label_no = 0;

  labelobj->link.type = ISEQ_ELEMENT_LABEL;
  labelobj->link.next = 0;

  labelobj->label_no = label_no++;
  labelobj->sc_state = 0;
  return labelobj;
}

static INSN *new_insn_core(yarv_iseq_t *iseqobj, int line_no,
                                         int insn_id, int argc, VALUE *argv){
  INSN *insnobj = compile_data_alloc_insn(iseqobj);
  
  insnobj->link.type = ISEQ_ELEMENT_INSN;
  insnobj->link.next = 0;
  insnobj->insn_id   = insn_id;
  insnobj->line_no   = line_no;
  insnobj->operands  = argv;
  insnobj->operand_size = argc;
  insnobj->sc_state = 0;
  return insn_optimize(insnobj);
}

static INSN *new_insn_body(yarv_iseq_t *iseqobj, int line_no,
                                         int insn_id, int argc, ...){
  VALUE *operands = 0;
  va_list argv;
  if(argc > 0){
    int i;
    va_init_list(argv, argc);
    operands = (VALUE *)compile_data_alloc(iseqobj, sizeof(VALUE) * argc);
    for(i=0; i<argc; i++){
      VALUE v = va_arg(argv, VALUE);
      operands[i] = v;

      /* TODO: need GC mark? */
      //if(!SPECIAL_CONST_P(v)){
      //  rb_ary_push(iseqobj->iseq_mark_ary, v);
      //}
    }
    va_end(argv);
  }
  return new_insn_core(iseqobj, line_no, insn_id, argc, operands);
}

static INSN *new_insn_send(yarv_iseq_t *iseqobj, int line_no,
                                         VALUE id, VALUE argc, VALUE block, VALUE flag){
  INSN *iobj = 0;
#ifdef OPT_BASIC_OPERATIONS
  if(block == 0 && flag == I2F(0)){
    ID mid = SYM2ID(id);

    if(argc == I2F(1)){
      if(0){
      }
      else if(mid == idPLUS){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_plus), 0);
      }
      else if(mid == idMINUS){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_minus), 0);
      }
      else if(mid == idMULT){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_mult), 0);
      }
      else if(mid == idDIV){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_div), 0);
      }
      else if(mid == idMOD){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_mod), 0);
      }
      else if(mid == idEq){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_eq), 0);
      }
      else if(mid == idLT){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_lt), 0);
      }
      else if(mid == idLTLT){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_ltlt), 0);
      }
      else if(mid == idAREF){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_aref), 0);
      }
    }
    else if(argc == I2F(0)){
      if(mid == idLength){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_length), 0);
      }
    }
    else if(argc == I2F(2)){
      if(mid == idASET){
        iobj = new_insn_body(iseqobj, line_no, BIN(opt_aset), 0);
      }
    }
  }
#endif
  if(iobj == 0){
    VALUE *operands = (VALUE *)compile_data_alloc(iseqobj, sizeof(VALUE) * 5);
    operands[0] = id;
    operands[1] = argc;
    operands[2] = block;
    operands[3] = flag;
    operands[4] = 0;
    iobj = new_insn_core(iseqobj, line_no, BIN(send), 5, operands);
  }
  return iobj;
}

static VALUE new_child_iseq(VALUE self, yarv_iseq_t *iseqobj,
                            NODE *node, VALUE name, VALUE parent, VALUE type){
  VALUE args[5];
  VALUE ret;
  
  debugs("[new_child_iseq]> ---------------------------------------\n");
  args[0] = (VALUE)node;
  args[1] = name;
  args[2] = iseq_filename(self);
  args[3] = parent;
  args[4] = type;
  
  ret = rb_class_new_instance(5, args, cYarvISeq);
  
  debugs("[new_child_iseq]< ---------------------------------------\n");
  rb_ary_push(iseqobj->compile_data->mark_ary, ret);
  return ret;
}

static int iseq_setup(VALUE self, LINK_ANCHOR *anchor){
  yarv_iseq_t *iseqobj;
  GetISeqVal(self, iseqobj);

//  debugs("[compile step 2] (iseq_array_to_linkedlist)\n");
//  iseq_array_to_linkedlist_each(anchor, seq_ary);
  
  GC_CHECK();
  if(CPDEBUG > 5)dump_disasm_list(FIRST_ELEMENT(anchor));
  GC_CHECK();

  debugs("[compile step 3.1 (iseq_optimize)]\n");
  iseq_optimize(iseqobj, anchor);

  if(CPDEBUG > 5)dump_disasm_list(FIRST_ELEMENT(anchor));
  GC_CHECK();
  
  debugs("[compile step 3.2 (iseq_insns_unification)]\n");
  iseq_insns_unification(iseqobj, anchor);
  if(CPDEBUG > 5)dump_disasm_list(FIRST_ELEMENT(anchor));
  GC_CHECK();
  
  debugs("[compile step 3.3 (set_sequence_stackcaching)]\n");
  set_sequence_stackcaching(iseqobj, anchor);
  if(CPDEBUG > 5)dump_disasm_list(FIRST_ELEMENT(anchor));
  GC_CHECK();
  
  debugs("[compile step 4 (set_sequence)]\n");
  set_sequence(iseqobj, anchor);
  if(CPDEBUG > 5)dump_disasm_list(FIRST_ELEMENT(anchor));
  GC_CHECK();

  GC_CHECK();
  debugs("[compile step 4.1 (set_exception_table)]\n");
  set_exception_table(iseqobj);

  debugs("[compile step 5] (set_optargs_table)\n");
  set_optargs_table(iseqobj);
  
  debugs("[compile step 6] (iseq_translate_direct_threaded_code)\n");
  iseq_translate_direct_threaded_code(iseqobj);
  GC_CHECK();
  
  if(CPDEBUG > 5){
    VALUE str = iseq_disasm(self);
    printf("%s\n", StringValueCStr(str));
  }
  debugs("[compile step: finish]\n");
  
  return 0;
}

VALUE iseq_assemble_setup(VALUE self, VALUE args, VALUE locals, VALUE insn_ary){
  yarv_iseq_t *iseqobj;
  int alen = RARRAY(args)->len, llen = RARRAY(locals)->len;
  int i;
  
  GetISeqVal(self, iseqobj);

  /* setup iseqobj (args, locals) */
  iseqobj->argc = RARRAY(args)->len;
  
  /* TODO: rest and opt should be supported */
  iseqobj->arg_simple = 1;

  if(iseqobj->local_tbl){
    ruby_xfree(iseqobj->local_tbl);
  }
  iseqobj->local_tbl  = (ID*)ALLOC_N(ID*, alen + llen);
  iseqobj->local_size = alen + llen;

  for(i=0; i<alen; i++){
    iseqobj->local_tbl[i]      = SYM2ID(RARRAY(args)->ptr[i]);
  }
  for(i=0; i<llen; i++){
    iseqobj->local_tbl[i+alen] = SYM2ID(RARRAY(locals)->ptr[i]);
  }
  
  iseqobj->catch_table_ary = rb_ary_new();
  // return iseq_setup(self, insn_ary, iseqobj);
  return Qnil;
}


int set_exception_tbl(yarv_iseq_t *iseqobj){
  static ID id_dollar_bang;

  if(!id_dollar_bang){
    id_dollar_bang = rb_intern("$!");
  }
  iseqobj->local_tbl = (ID*)ALLOC_N(ID*, 1);
  iseqobj->local_size = 1;
  iseqobj->local_tbl[0] = id_dollar_bang;
  return COMPILE_OK;
}


NODE* set_block_local_tbl(VALUE self, yarv_iseq_t *iseqobj, NODE *node, LINK_ANCHOR *anchor){
  VALUE vars = rb_ary_new();
  VALUE tmp;
  NODE *nargs, *ntmp, *nelem;
  int dlocal = 0;
  
  /* argument check */
  if(iseqobj->type != ISEQ_TYPE_BLOCK){
    rb_bug("set_block_local_tbl: unexpected iseq type");
  }

  /* This code will be vanished                                     */
  /* because variable appeared in block first will be method local. */
  /* patch from Kent Sibilev. Thanks.                               */
  if(node && nd_type(node) == NODE_BLOCK){
    nelem = node->nd_head;
    while(nelem && nd_type(nelem) == NODE_DASGN_CURR){
      if(!rb_ary_includes(vars, tmp = I2F(nelem->nd_vid))){
        dlocal++;
      }
      nelem = nelem->nd_value;
    }
  }

  
  nargs = node->nd_var;
  if(nargs){
    switch(nd_type(nargs)){
    case NODE_MASGN:
      if(nargs->nd_head == 0){
        rb_bug("unsupport this type block parameter");
      }
      else{
        iseqobj->argc = nargs->nd_head->nd_alen;
        ntmp = nargs->nd_head;
        while(ntmp){
          // printf("-- %s\n", node_name(nd_type(ntmp)));
          nelem = ntmp->nd_head;
          if(nd_type(nelem) != NODE_DASGN_CURR){
            char buff[0x20];
            ID   id;
            int  idx = iseqobj->argc - RARRAY(vars)->len;
            DECL_ANCHOR(anc);
            
            snprintf(buff, 0x20, "__tp%d__", idx);
            id = rb_intern(buff);
            /* idx-th param, current level*/
            ADD_INSN2(anchor, nd_line(node), getdynamic, I2F(idx + dlocal), I2F(0));
            COMPILE(anc, "set_block_local_tbl#masgn/other", nelem);
            ADD_ELEM(anchor, LAST_ELEMENT(anc));
            rb_ary_push(vars, I2F(id));
          }
          else{
            rb_ary_push(vars, I2F(nelem->nd_vid));
            debugi("set_block_local_tbl#NODE_MASGN", (nelem->nd_vid));
          }
          ntmp = ntmp->nd_next;
        }
      }
      break;
      
    case NODE_DASGN_CURR:
      iseqobj->argc = 1;
      rb_ary_push(vars, I2F(nargs->nd_vid));
      debugi("set_block_local_tbl#NODE_DASGN_CURR", (nargs->nd_vid));
      break;

    /* for 1.x compatibility */
    default:{
      /* initial values */
      static ID tid;
      DECL_ANCHOR(anc);
      
      if(tid == 0){
        tid = rb_intern("__tp__");
      }
      iseqobj->argc = 1;
      ADD_INSN2(anchor, nd_line(node), getdynamic, I2F(1), I2F(0)); /* first param, current level*/
      COMPILE(anc, "set_block_local_tbl#lasgn/dasgn", nargs);
      ADD_ELEM(anchor, POP_ELEMENT(anc));
      rb_ary_push(vars, I2F(tid));
      debugi("set_block_local_tbl#lasgin/dasgn", (nargs->nd_vid));
      break;
    }
    }
  }
  else{
    iseqobj->argc = 0;
  }
  iseqobj->arg_rest = 0;
  node = node->nd_body;

  /* block local variables check */
  /* this part will be deleted   */
  while(node != 0 && nd_type(node) == NODE_BLOCK){
    nelem = node->nd_head;
    if(nelem == 0                         ||
       nd_type(nelem)  != NODE_DASGN_CURR ||
       nelem->nd_value != 0               ||
       rb_ary_includes(vars, tmp = I2F(nelem->nd_vid))){
      break;
    }
    rb_ary_push(vars, tmp);
    node = node->nd_next;
  }
  
  /* This code will be vanished                                     */
  /* because variable appeared in block first will be method local. */
  /* patch from Kent Sibilev. Thanks.                               */
  if(node && nd_type(node) == NODE_BLOCK){
    nelem = node->nd_head;
    while(nelem && nd_type(nelem) == NODE_DASGN_CURR){
      if(!rb_ary_includes(vars, tmp = I2F(nelem->nd_vid))){
        rb_ary_push(vars, tmp);
      }
      nelem = nelem->nd_value;
    }
  }

  {
    int i, size = RARRAY(vars)->len;
    
    if(size > 0){
      iseqobj->local_tbl = (ID*)ALLOC_N(ID*, size);
      for(i=0; i<size; i++){
        iseqobj->local_tbl[i] = FIX2INT(RARRAY(vars)->ptr[i]);
        debugi("block param", iseqobj->local_tbl[i]);
      }
    }
    iseqobj->local_size = size;
  }
  return node;
}

static int get_dyna_var_idx_at_raw(yarv_iseq_t *iseqobj, ID id){
  int i;
  for(i=0; i<iseqobj->local_size; i++){
    if(iseqobj->local_tbl[i] == id){
      return i;
    }
  }
  return -1;
}

static int get_dyna_var_idx_at(VALUE self, ID id){
  yarv_iseq_t *iseqobj;
  GetISeqVal(self, iseqobj);
  return get_dyna_var_idx_at_raw(iseqobj, id);
}

static int get_dyna_var_idx(VALUE self, ID id, int *level, int *ls){
  yarv_iseq_t *iseqobj;
  int lv = 0, idx;
  GetISeqVal(self, iseqobj);
  
  while(iseqobj){
    if((idx = get_dyna_var_idx_at_raw(iseqobj, id)) >= 0){
      *level = lv;
      *ls    = iseqobj->local_size;
      return idx;
    }
    iseqobj = iseqobj->parent_iseqobj;
    lv++;
  }
  return -1;
}

/**

 */
int set_arguments(VALUE self, yarv_iseq_t *iseqobj, LINK_ANCHOR *optargs, NODE *node){
  int i, j;
  
  if(node){
    /* normal method */
    if(node->nd_frml){
      iseqobj->argc = RARRAY(node->nd_frml)->len;
    }
    else{
      iseqobj->argc = 0;
    }

    /* TODO */
    if(node->nd_rest){
      iseqobj->arg_rest = node->nd_rest->nd_cnt - 2;
    }
    else{
      iseqobj->arg_rest = -1;
    }

    /* optional initializer */
    if(node->nd_opt){
      NODE* optarg = node->nd_opt;
      LABEL* label;
      VALUE labels = rb_ary_new();
      i = 0;
      while(optarg){
        label = NEW_LABEL(nd_line(node));
        rb_ary_push(labels, (VALUE) label | 1);
        ADD_LABEL(optargs, label);
        COMPILE_POPED(optargs, "optarg", optarg->nd_head);

        optarg = optarg->nd_next;
        i += 1;
      }
      /* last label */
      label = NEW_LABEL(nd_line(node));
      rb_ary_push(labels,  (VALUE) label | 1);
      ADD_LABEL(optargs, label);
      i+=1;
      
      iseqobj->arg_opts = i;
      iseqobj->arg_opt_tbl = ALLOC_N(VALUE, i);
      MEMCPY(iseqobj->arg_opt_tbl, RARRAY(labels)->ptr, VALUE, i);
      for(j=0; j<i; j++){
        iseqobj->arg_opt_tbl[j] &= ~1;
      }
    }
    else{
      iseqobj->arg_opts = 0;
    }
  }
  else{
    iseqobj->argc     =  0;
    iseqobj->arg_rest = -1;
    iseqobj->arg_opts =  0;
  }

  if(iseqobj->arg_rest + 1 != 0 ||
     iseqobj->arg_opts     != 0){
    iseqobj->arg_simple = 0;
  }
  else{
    iseqobj->arg_simple = 1;
  }
  return COMPILE_OK;
}


/**

 */
int set_localtbl(yarv_iseq_t *iseqobj, ID* tbl){
  int size;
  if(tbl){
    size = *tbl - 2 /* $~, $_ */ + 1 /* svar location */;
  }
  else{
    size = 1;
  }
  iseqobj->local_tbl = (ID*)ALLOC_N(ID*, size);
  if(tbl && size > 1){
    MEMCPY(iseqobj->local_tbl + 1, tbl + 3, ID*, size - 1);
  }
  iseqobj->local_size = size;
  return COMPILE_OK;
}

/**
  ruby insn object array -> raw instruction sequence
 */
static int set_sequence(yarv_iseq_t *iseqobj, LINK_ANCHOR *anchor){
  LABEL *lobj;
  INSN  *iobj;
  struct insn_info_struct *insn_info_tbl;
  LINK_ELEMENT *list;
  VALUE *generated_iseq;
  
  int k, pos, stack_max = 0;

  GC_CHECK();
  /* set label position */
  list = FIRST_ELEMENT(anchor);
  k = pos = 0;
  while(list){
    switch(list->type){
    case ISEQ_ELEMENT_INSN:{
      iobj = (INSN*)list;
      pos += insn_data_length(iobj);
      k+=1;
      break;
    }
    case ISEQ_ELEMENT_LABEL:{
      lobj = (LABEL*)list;
      lobj->position = pos;
      lobj->set      = Qtrue;
      break;
    }
    case ISEQ_ELEMENT_NONE:{
      /* ignore */
      break;
    }
    default:
      dump_disasm_list(FIRST_ELEMENT(anchor));
      dump_disasm_list(list);
      rb_bug("error: set_sequence");
      break;
    }
    list = list->next;
  }

  /* make instruction sequence */
  generated_iseq = ALLOC_N(VALUE, pos);
  insn_info_tbl = ALLOC_N(struct insn_info_struct, k);

  GC_CHECK();
  
  list = FIRST_ELEMENT(anchor);
  k = pos  = 0;
  
  while(list){
    switch(list->type){
    case ISEQ_ELEMENT_INSN:{
      int j, len, insn;
      char *types;
      VALUE *operands;
      
      iobj = (INSN*)list;
      operands = iobj->operands;
      insn = iobj->insn_id;
      generated_iseq[pos] = insn;
      types = insn_op_types(insn);
      len   = insn_len(insn);
      stack_max += insn_ret_num(insn);

      /* operand check */
      if(iobj->operand_size != len - 1){
        dump_disasm_list(list);
        rb_bug("operand size miss! (%d for %d)", iobj->operand_size, len - 1);
        return 0;
      }

      for(j=0; types[j]; j++){
        char type = types[j];
        // printf("--> [%c - (%d-%d)]\n", type, k, j);
        switch(type){
        case 'L':{          /* label(destination position) */
          lobj = (LABEL*)operands[j];
          if(lobj->set != Qtrue){
            rb_bug("unknown label");
          }
          generated_iseq[pos+1+j] = lobj->position - (pos + len);
          break;
        }
        case 'H':{
          /*
           * [[obj, label], ...]
           */
          int i;
          VALUE lits = ((NODE*)operands[j])->u1.value;
          VALUE map  = rb_hash_new();

          for(i=0; i<RARRAY(lits)->len; i++){
            VALUE pair = RARRAY(lits)->ptr[i];
            lobj = (LABEL *)(RARRAY(pair)->ptr[1] & ~1);
            
            if(lobj->set != Qtrue){
              rb_bug("unknown label");
            }
            rb_hash_aset(map, RARRAY(pair)->ptr[0], INT2FIX(lobj->position - (pos + len)));
          }
          generated_iseq[pos+1+j] = map;
          rb_ary_push(iseqobj->iseq_mark_ary, map);
          break;
        }
        case 'U':          /* ulong */
          generated_iseq[pos+1+j] = FIX2INT(operands[j]);
          break;
        case 'V':          /* VALUE */
        case 'C':          /* class path */
          {
            VALUE v = operands[j];
            generated_iseq[pos+1+j] = v;
            /* to mark ruby object */
            if(!SPECIAL_CONST_P(v)){
              rb_ary_push(iseqobj->iseq_mark_ary, v);
            }
          }
          break;
        case 'M':          /* inline cache */
          {
            VALUE v = (VALUE)NEW_INLINE_CACHE_ENTRY();
            generated_iseq[pos+1+j] = v;
            rb_ary_push(iseqobj->iseq_mark_ary, v);
            break;
          }
        case 'I':          /* ID */
          generated_iseq[pos+1+j] = SYM2ID(operands[j]);
          break;
        case 'G':
          {
            struct global_entry *entry = (struct global_entry*)(operands[j] & (~1));
            generated_iseq[pos+1+j] = (VALUE)entry;
          }
          break;
        default:
          rb_bug("unknown operand type: %c", type);
          return 0;
        }
      }
      insn_info_tbl[k].line_no  = iobj->line_no;
      insn_info_tbl[k].position = pos;
      pos += len;
      k++;
      break;
    }
    default:
      /* ignore */
      break;
    }
    list = list->next;
  }

  {
    iseqobj->iseq = (void*)generated_iseq;
    iseqobj->size  = pos;
    iseqobj->insn_info_tbl = insn_info_tbl;
    iseqobj->insn_info_size= k;
    iseqobj->stack_max = stack_max;
  }
  return COMPILE_OK;
}

static int label_get_position(VALUE self){
  LABEL *lobj = (LABEL *) self;
  return lobj->position;
}

int set_exception_table(yarv_iseq_t *iseqobj){
  VALUE *tptr, *ptr;
  int tlen, i;
  struct catch_table_entry *entry;

  tlen = RARRAY(iseqobj->catch_table_ary)->len;
  tptr = RARRAY(iseqobj->catch_table_ary)->ptr;

  iseqobj->catch_table = ALLOC_N(struct catch_table_entry, tlen);
  iseqobj->catch_table_size = tlen;
  
  for(i=0; i<tlen; i++){
    ptr   = RARRAY(tptr[i])->ptr;
    entry = &iseqobj->catch_table[i];
    entry->type  = ptr[0];
    entry->start = label_get_position(ptr[1] & ~1);
    entry->end   = label_get_position(ptr[2] & ~1);
    entry->iseq  = ptr[3];
    
    if(entry->iseq != 0){
      rb_ary_push(iseqobj->iseq_mark_ary, entry->iseq);
    }
    
    entry->sp    = FIX2INT(ptr[4]);
    if(ptr[5]){
      entry->cont  = label_get_position(ptr[5] & ~1);
    }
    else{
      entry->cont  = 0;
    }
  }
  //
  iseqobj->catch_table_ary = 0; /* free */
  return COMPILE_OK;
}

/*
 * set optional argument table
 *   def foo(a, b=expr1, c=expr2)
 *   => 
 */
int set_optargs_table(yarv_iseq_t *iseqobj){
  int i;

  if(iseqobj->arg_opts != 0){
    for(i=0; i< iseqobj->arg_opts; i++){
      // dp((VALUE)iseqobj->arg_opt_tbl[i]);
      iseqobj->arg_opt_tbl[i] =
        label_get_position((VALUE)iseqobj->arg_opt_tbl[i]);
    }
  }
  return COMPILE_OK;
}

static LINK_ELEMENT *
get_destination_insn(INSN *iobj){
  LABEL *lobj = (LABEL *)OPERAND_AT(iobj, 0);
  LINK_ELEMENT *list;
  
  list = lobj->link.next;
  while(list){
    if(list->type == ISEQ_ELEMENT_INSN){
      break;
    }
    list = list->next;
  }
  return list;
}

static LINK_ELEMENT *
get_next_insn(INSN *iobj){
  LINK_ELEMENT *list = iobj->link.next;
  
  while(list){
    if(list->type == ISEQ_ELEMENT_INSN){
      return list;
    }
    list = list->next;
  }
  return 0;
}

static LINK_ELEMENT *
get_prev_insn(INSN *iobj){
  LINK_ELEMENT *list = iobj->link.prev;
  
  while(list){
    if(list->type == ISEQ_ELEMENT_INSN){
      return list;
    }
    list = list->prev;
  }
  return 0;
}

static int iseq_optimize(yarv_iseq_t *iseqobj, LINK_ANCHOR *anchor){
  LINK_ELEMENT *list;
  /*
   *  useless jump elimination:
   *     jump LABEL1
   *     ...
   *   LABEL2:
   *     jump LABEL2
   *
   *   => in this case, first jump instruction should jump tp
   *      LABEL2 directly
   */
  list = FIRST_ELEMENT(anchor);
  // dump_disasm_list(list);
  while(list){
    if(list->type == ISEQ_ELEMENT_INSN){
      INSN *niobj, *diobj, *piobj, *iobj = (INSN *)list;
      if(iobj->insn_id == BIN(jump)){
        diobj = (INSN*)get_destination_insn(iobj);
        niobj = (INSN*)get_next_insn(iobj);
        
        if(diobj == niobj){
          REMOVE_ELEM(&iobj->link);
        }
        else if(diobj->insn_id == BIN(jump)){
          OPERAND_AT(iobj, 0) = OPERAND_AT(diobj, 0);
          continue;
        }
        else if(diobj->insn_id == BIN(end)){
          INSN *eiobj = new_insn_core(iseqobj, iobj->line_no, BIN(end),
                                                    diobj->operand_size, diobj->operands);
          /* replace */
          REPLACE_ELEM((LINK_ELEMENT *)iobj,
                       (LINK_ELEMENT *)eiobj);
        }
        /*
         *   if   L1
         *   jump L2
         * L1:
         *   ...
         * L2:
         *
         * ==>
         *   unless L2
         * L1:
         *   ...
         * L2:
         */
        else if ((piobj = (INSN*)get_prev_insn(iobj)) != 0 &&
                 (piobj->insn_id == BIN(if) ||
                  piobj->insn_id == BIN(unless))){
          if(niobj == (INSN*)get_destination_insn(piobj)){
            piobj->insn_id = (piobj->insn_id == BIN(if)) ? BIN(unless) : BIN(if);
            OPERAND_AT(piobj, 0) = OPERAND_AT(iobj, 0);
            REMOVE_ELEM(&iobj->link);
          }
        }
      }

      if(iobj->insn_id == BIN(end)){
        INSN *piobj = (INSN*)get_prev_insn((INSN*)list);
        if(piobj->insn_id == BIN(send)){
          /* TODO: tail call optimization */
          if(piobj->operands[2] == 0){
            //piobj->operands[3] = INT2FIX(FIX2INT(piobj->operands[3]) | VM_CALL_TAILCALL_BIT);
            //piobj->operands[3] = INT2FIX(FIX2INT(piobj->operands[3]) | VM_CALL_TAILRECURSION_BIT);
          }
        }
      }
    }
    list = list->next;
  }
  return COMPILE_OK;
}

static INSN* new_unified_insn(yarv_iseq_t *iseqobj,
                                     int insn_id, int size, LINK_ELEMENT *seq_list){
  INSN *iobj = 0;
  LINK_ELEMENT *list = seq_list;
  int i, argc = 0;
  VALUE *operands = 0, *ptr = 0;

  
  /* count argc */
  for(i=0; i<size; i++){
    iobj = (INSN*)list;
    argc += iobj->operand_size;
    list = list->next;
  }

  if(argc > 0){
    ptr = operands = (VALUE *)compile_data_alloc(iseqobj, sizeof(VALUE) * argc);
  }
  
  /* copy operands */
  list = seq_list;
  for(i=0; i<size; i++){
    iobj = (INSN*)list;
    MEMCPY(ptr, iobj->operands, VALUE, iobj->operand_size);
    ptr += iobj->operand_size;
    list = list->next;
  }
  
  return new_insn_core(iseqobj, iobj->line_no, insn_id, argc, operands);
}


/*
 * This scheme can get more performance if do this optimize with
 * label address resolving.
 * It's future work (if compile time was bottle neck).
 */
static int iseq_insns_unification(yarv_iseq_t *iseqobj, LINK_ANCHOR *anchor){
#ifdef OPT_INSTRUCTIONS_UNIFICATION
  LINK_ELEMENT *list;
  INSN *iobj, *niobj;
  int id, j, k;
  
  list = FIRST_ELEMENT(anchor);
  while(list){
    if(list->type == ISEQ_ELEMENT_INSN){
      iobj = (INSN*)list;
      id = iobj->insn_id;
      if(unified_insns_data[id] != 0){
        int **entry = unified_insns_data[id];
        for(j=1; j<(int)entry[0]; j++){
          int *unified = entry[j];
          LINK_ELEMENT *li = list->next;
          for(k=2; k<unified[1]; k++){
            if(li->type != ISEQ_ELEMENT_INSN ||
               ((INSN *)li)->insn_id != unified[k]){
              goto miss;
            }
            li = li->next;
          }
          /* matched */
          niobj = new_unified_insn(iseqobj, unified[0], unified[1]-1, list);

          /* insert to list */
          niobj->link.prev = (LINK_ELEMENT *)iobj->link.prev;
          niobj->link.next = li;
          if(li){
            li->prev = (LINK_ELEMENT *)niobj;
          }

          list->prev->next = (LINK_ELEMENT *)niobj;
          list = (LINK_ELEMENT *)niobj;
          break;
        miss:;
       }
      }
    }
    list = list->next;
  }
#endif
  return COMPILE_OK;
}

#ifdef OPT_STACK_CACHING

#define SC_INSN(insn, stat) sc_insn_info[(insn)][(stat)]
#define SC_NEXT(insn)       sc_insn_next[insn]

#include "opt_sc.inc"

int insn_set_sc_state(INSN *iobj, int state){
  int nstate;
  int insn_id;

  insn_id       = iobj->insn_id;
  iobj->insn_id = SC_INSN(insn_id, state);
  nstate        = SC_NEXT(iobj->insn_id);

  if(insn_id == BIN(jump) ||
     insn_id == BIN(if)   ||
     insn_id == BIN(unless)){
    LABEL *lobj = (LABEL*) OPERAND_AT(iobj, 0);

    if(lobj->sc_state != 0){
      if(lobj->sc_state != nstate){
        dump_disasm_list((LINK_ELEMENT*) iobj);
        printf("%d, %d: ", lobj->sc_state, nstate);
        rb_bug("insn_set_sc_state error\n");
        return 0;
      }
    }
    else{
      lobj->sc_state = nstate;
    }
    if(insn_id == BIN(jump)){
      nstate = SCS_XX;
    }
  }
  else if(insn_id == BIN(end)){
    nstate = SCS_XX;
  }
  
  return nstate;
}

int label_set_sc_state(LABEL *lobj, int state){
  if(lobj->sc_state != 0){
    if(lobj->sc_state != state){
      state = lobj->sc_state;
    }
  }
  else{
    lobj->sc_state = state;
  }
  
  return state;
}


#endif

static int set_sequence_stackcaching(yarv_iseq_t *iseqobj, LINK_ANCHOR *anchor){
#ifdef OPT_STACK_CACHING
  LINK_ELEMENT *list;
  int state, insn_id;

  /* initialize */
  state = SCS_XX;
  list = FIRST_ELEMENT(anchor);

  /* for each list element */
  while(list){
  redo_point:
    switch(list->type){
    case ISEQ_ELEMENT_INSN:{
      INSN *iobj = (INSN *)list;
      insn_id = iobj->insn_id;

      // dump_disasm_list(list);

      switch(insn_id){
      case BIN(nop):{
        /* exception merge point */
        if(state != SCS_AX){
          INSN *rpobj = new_insn_body(iseqobj, 0, BIN(reput), 0);
          
          /* replace this insn */
          REPLACE_ELEM(list, (LINK_ELEMENT *)rpobj);
          list = (LINK_ELEMENT *)rpobj;
          goto redo_point;
        }
        break;
      }
      case BIN(swap):{
        if(state == SCS_AB || state == SCS_BA){
          state = (state == SCS_AB ? SCS_BA : SCS_AB);
          
          REMOVE_ELEM(list);
          list = list->next;
          goto redo_point;
        }
        break;
      }
      case BIN(pop):{
        switch(state){
        case SCS_AX:
        case SCS_BX:
          state = SCS_XX;
          break;
        case SCS_AB:
          state = SCS_AX;
          break;
       case SCS_BA:
          state = SCS_BX;
          break;
        case SCS_XX:
          goto normal_insn;
        default:
          rb_bug("unreachable");
        }
        /* remove useless pop */
        REMOVE_ELEM(list);
        list = list->next;
        goto redo_point;
      }
      default:;
        /* none */
      } /* end of switch */
    normal_insn:
      state = insn_set_sc_state(iobj, state);
      break;
    }
    case ISEQ_ELEMENT_LABEL:{
      LABEL *lobj;
      lobj = (LABEL *)list;

      state = label_set_sc_state(lobj, state);
    }
    default:
      break;
    }
    list = list->next;
  }
#endif
  return COMPILE_OK;
}



int compile_dstr(VALUE self, yarv_iseq_t *iseqobj, LINK_ANCHOR *ret, NODE *node){
  NODE *list = node->nd_next;
  VALUE lit  = node->nd_lit;
  int cnt = 0;

  debugp_param("nd_lit", lit);
  if(RSTRING(lit)->len == 0){
    cnt = 0;
  }
  else{
    ADD_INSN1(ret, nd_line(node), putobject, node->nd_lit);
    cnt = 1;
  }

  while(list){
    COMPILE(ret, "each string", list->nd_head);
    cnt++;
    list = list->nd_next;
  }
  if(cnt > 1){
    ADD_INSN1(ret, nd_line(node), concatstrings, I2F(cnt));
  }
  
  return COMPILE_OK;
}

static yarv_iseq_t *get_root_iseq_object(yarv_iseq_t *iop){
  yarv_iseq_t *iseqobj = iop;
  while(iseqobj){
    if(iseqobj->parent_iseqobj){
      iseqobj = iseqobj->parent_iseqobj;
    }
    else{
      return iseqobj;
    }
  }
  /* unreachable */
  rb_bug("get_root_iseq_object error");
  return 0;
}

static int get_root_iseq_localsize(yarv_iseq_t *iseqobj){
  iseqobj = get_root_iseq_object(iseqobj);
  return iseqobj->local_size;
}

static int compile_branch_condition(VALUE self, yarv_iseq_t *iseqobj,
                                    LINK_ANCHOR *ret, NODE *cond,
                                    LABEL *then_label, LABEL *else_label){
  
  switch(nd_type(cond)){
  case NODE_NOT:
    compile_branch_condition(self, iseqobj, ret, cond->nd_body, else_label, then_label);
    break;

  case NODE_AND:{
    LABEL *label = NEW_LABEL(nd_line(cond));
    compile_branch_condition(self, iseqobj, ret, cond->nd_1st, label, else_label);
    ADD_LABEL(ret, label);
    compile_branch_condition(self, iseqobj, ret, cond->nd_2nd, then_label, else_label);
    break;
  }
  case NODE_OR:{
    LABEL *label = NEW_LABEL(nd_line(cond));
    compile_branch_condition(self, iseqobj, ret, cond->nd_1st, then_label, label);
    ADD_LABEL(ret, label);
    compile_branch_condition(self, iseqobj, ret, cond->nd_2nd, then_label, else_label);
    break;
  }
  default:
    COMPILE(ret, "branch condition", cond);
    ADD_INSNL(ret, nd_line(cond), unless, else_label);
    ADD_INSNL(ret, nd_line(cond), jump,   then_label);
    break;
  }
  return COMPILE_OK;
}

static int compile_array(VALUE self, yarv_iseq_t *iseqobj,
                         LINK_ANCHOR *ret, NODE *node_root, VALUE opt_p){
  NODE *node = node_root;
  int len = node->nd_alen, line = nd_line(node);
  DECL_ANCHOR(anchor);
  
  while(node){
    if(opt_p && nd_type(node->nd_head) != NODE_LIT){
      opt_p = Qfalse;
    }
    COMPILE(anchor, "element", node->nd_head);
    node = node->nd_next;
  }

  if(opt_p == Qtrue){
    VALUE ary = rb_ary_new();
    node = node_root;
    while(node){
      rb_ary_push(ary, node->nd_head->nd_lit);
      node = node->nd_next;
    }

    rb_ary_push(iseqobj->compile_data->mark_ary, ary);
    ADD_INSN1(ret, nd_line(node_root), duparray, ary);
  }
  else{
    ADD_INSN1(anchor, line, newarray, I2F(len));
    APPEND_LIST(ret, anchor);
  }
  
  return COMPILE_OK;
}

static VALUE case_when_optimizable_literal(NODE *node){
  if(nd_type(node) == NODE_LIT){
    VALUE v     = node->nd_lit;
    VALUE klass = CLASS_OF(v);
    if(klass == rb_cSymbol || rb_obj_is_kind_of(v, rb_cNumeric)){
      return v;
    }
  }
  else if(nd_type(node) == NODE_STR){
    return node->nd_lit;
  }
  return Qfalse;
}

static int make_masgn_lhs(VALUE self, yarv_iseq_t *iseqobj,
                          LINK_ANCHOR *ret, NODE *node){
  
  switch(nd_type(node)){
  case NODE_ATTRASGN:{
    INSN *iobj;
    VALUE dupidx;

    COMPILE(ret, "masgn lhs (NODE_ATTRASGN)", node);
    iobj = (INSN*)ret->last;
    ret->last = ret->last->prev;
    
    dupidx = iobj->operands[1]; // RARRAY(iobj->operands)->ptr[1]; /* send sym, num, ... */
    dupidx = I2F(FIX2INT(dupidx) + 1);
    iobj->operands[1] = dupidx;

    ADD_INSN1(ret, nd_line(node), topn, dupidx);
    ADD_ELEM(ret, (LINK_ELEMENT*) iobj);
    ADD_INSN (ret, nd_line(node), pop); /* result */
    ADD_INSN (ret, nd_line(node), pop); /* rhs    */
    break;
  }

  default:{
    DECL_ANCHOR(anchor);
    
    COMPILE_POPED(anchor, "masgn lhs", node);
    ADD_ELEM(ret, LAST_ELEMENT(anchor));
  }
  }
  
  return COMPILE_OK;
}

static int defined_expr(VALUE self, yarv_iseq_t *iseqobj,
                        LINK_ANCHOR *ret, NODE *node, LABEL *lfinish, VALUE needstr){
  char *estr = 0;
  
  switch(nd_type(node)){

  /* easy literals */
  case NODE_NIL:
    estr = "nil";
    break;
  case NODE_SELF:
    estr = "self";
    break;
  case NODE_TRUE:
    estr = "true";
    break;
  case NODE_FALSE:
    estr = "false";
    break;

  /* variables */
  case NODE_LVAR:
    estr = "local-variable";
    break;
  case NODE_DVAR:
    estr = "local-variable(in-block)";
    break;

  case NODE_IVAR:
    ADD_INSN3(ret, nd_line(node), defined, I2F(DEFINED_IVAR), ID2SYM(node->nd_vid), needstr);
    return 1;
  case NODE_GVAR:
    ADD_INSN3(ret, nd_line(node), defined, I2F(DEFINED_GVAR), ((VALUE)node->nd_entry) | 1, needstr);
    return 1;
  case NODE_CVAR:
    ADD_INSN3(ret, nd_line(node), defined, I2F(DEFINED_CVAR), ID2SYM(node->nd_vid), needstr);
    return 1;

  case NODE_CONST:
    ADD_INSN (ret, nd_line(node), putnil);
    ADD_INSN3(ret, nd_line(node), defined, I2F(DEFINED_CONST), ID2SYM(node->nd_vid), needstr);
    return 1;
  case NODE_COLON2:
    if(rb_is_const_id(node->nd_mid)){
      LABEL *lcont = NEW_LABEL(nd_line(node));
      defined_expr(self, iseqobj, ret, node->nd_head, lfinish, Qfalse);
      
      ADD_INSNL(ret, nd_line(node), if, lcont);
      ADD_INSN (ret, nd_line(node), putnil);
      ADD_INSNL(ret, nd_line(node), jump, lfinish);
      
      ADD_LABEL(ret, lcont);
      COMPILE(ret, "defined/colon2#nd_head", node->nd_head);
      ADD_INSN3(ret, nd_line(node), defined, I2F(DEFINED_CONST), ID2SYM(node->nd_mid), needstr);
    }
    else{
      LABEL *lcont = NEW_LABEL(nd_line(node));
      defined_expr(self, iseqobj, ret, node->nd_head, lfinish, Qfalse);
      
      ADD_INSNL(ret, nd_line(node), if, lcont);
      ADD_INSN (ret, nd_line(node), putnil);
      ADD_INSNL(ret, nd_line(node), jump, lfinish);
      
      ADD_LABEL(ret, lcont);
      COMPILE(ret, "defined/colon2#nd_head", node->nd_head);
      ADD_INSN3(ret, nd_line(node), defined, I2F(DEFINED_METHOD), ID2SYM(node->nd_mid), needstr);
    }
    return 1;
  case NODE_COLON3:
    ADD_INSN1(ret, nd_line(node), putobject, rb_cObject);
    ADD_INSN3(ret, nd_line(node), defined, I2F(DEFINED_CONST), ID2SYM(node->nd_mid), needstr);
    return 1;


  /* method dispatch */
  case NODE_CALL:
  case NODE_VCALL:
  case NODE_FCALL:
    if(nd_type(node) == NODE_CALL){
      LABEL *lcont = NEW_LABEL(nd_line(node));
      
      defined_expr(self, iseqobj, ret, node->nd_recv, lfinish, Qfalse);
      ADD_INSNL(ret, nd_line(node), if, lcont);
      ADD_INSN (ret, nd_line(node), putnil);
      ADD_INSNL(ret, nd_line(node), jump, lfinish);
      
      ADD_LABEL(ret, lcont);
      COMPILE(ret, "defined/recv", node->nd_recv);
    }
    else{
      ADD_INSN(ret, nd_line(node), putself);
    }
    ADD_INSN3(ret, nd_line(node), defined, I2F(DEFINED_METHOD), ID2SYM(node->nd_mid), needstr);
    return 1;
  default:
    rb_bug("unimplemented defined");
  }

  if(estr != 0){
    if(needstr != Qfalse){
      VALUE str = rb_str_new2(estr);
      ADD_INSN1(ret, nd_line(node), putstring, str);
      rb_ary_push(iseqobj->compile_data->mark_ary, str);
    }
    else{
      ADD_INSN1(ret, nd_line(node), putobject, Qtrue);
    }
    return 1;
  }
  return 0;
}

#define BUFSIZE 0x100

static
VALUE make_name_for_block(yarv_iseq_t *iseq){
  char buf[BUFSIZE];
  if(iseq->parent_iseqobj == 0){
    snprintf(buf, BUFSIZE, "block in %s", RSTRING(iseq->name)->ptr);
  }
  else{
    int level = 1;
    yarv_iseq_t *ip = iseq;
    while(1){
      if(ip->parent_iseqobj){
        ip = ip->parent_iseqobj;
      }
      else{
        break;
      }
      level++;
    }
    snprintf(buf, BUFSIZE, "block (%d levels) in %s", level,
             RSTRING(ip->name)->ptr);
  }
  return rb_str_new2(buf);
}

static
VALUE make_name_with_str(char *fmt, char *str){
  char buf[BUFSIZE];
  snprintf(buf, BUFSIZE, fmt, str);
  return rb_str_new2(buf);
}


static
void add_ensure_iseq(LINK_ANCHOR *ret, yarv_iseq_t *iseq, VALUE self){
  struct iseq_compile_data_ensure_node_stack *enlp = iseq->compile_data->ensure_node_stack;
  DECL_ANCHOR(ensure);

  while(enlp){
    DECL_ANCHOR(ensure_part);
    LINK_ANCHOR *t;
    COMPILE_POPED(ensure_part, "ensure part", enlp->ensure_node);

    ADD_SEQ(ensure, ensure_part);
    enlp = enlp->prev;
  }
  ADD_SEQ(ret, ensure);
}

/**
  compile each node

  self:  InstructionSequence
  node:  Ruby compiled node
  poped: This node will be poped
 */
static int iseq_compile_each(VALUE self, LINK_ANCHOR *ret, NODE* node, int poped){
  VALUE tmp; /* reserved for macro */
  
  int type;
  yarv_iseq_t *iseqobj;

  GetISeqVal(self, iseqobj);
  
  GC_CHECK();

  if(node == 0){
    if(!poped){
      debug_nodeprint("NODE_NIL(implicit)");
      debug_nodeprint_close();
      ADD_INSN(ret, 0, putnil);
      return COMPILE_OK;
    }
    return COMPILE_OK;
  }

  debug_nodeprint(node_name(nd_type(node)));
  
  type = nd_type(node);
  
  switch(type){

  case NODE_METHOD:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_METHOD"));
    break;
  }
  case NODE_FBODY:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_FBODY"));
    break;
  }
  case NODE_CFUNC:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_CFUNC"));
    break;
  }
  case NODE_SCOPE:{
    /* OK */
    COMPILE_ERROR(("BUG: shouldn't reach: NODE_SCOPE"));
    break;
  }
  case NODE_BLOCK:{
    while(node && nd_type(node) == NODE_BLOCK){
      COMPILE_(ret, "BLOCK body", node->nd_head,
               (node->nd_next == 0 && poped == 0) ? 0 : 1);
      node = node->nd_next;
    }
    if(node){
      COMPILE_(ret, "BLOCK next", node->nd_next, poped);
    }
    break;
  }
  case NODE_IF:{
    DECL_ANCHOR(cond_seq);
    DECL_ANCHOR(then_seq);
    DECL_ANCHOR(else_seq);
    LABEL *then_label, *else_label, *end_label;

    then_label = NEW_LABEL(nd_line(node));
    else_label = NEW_LABEL(nd_line(node));
    end_label  = NEW_LABEL(nd_line(node));

    compile_branch_condition(self, iseqobj, cond_seq, node->nd_cond, then_label, else_label);
    COMPILE_(then_seq, "then", node->nd_body, poped);
    COMPILE_(else_seq, "else", node->nd_else, poped);

    ADD_SEQ  (ret, cond_seq);
    
    ADD_LABEL(ret, then_label);
    ADD_SEQ  (ret, then_seq);
    ADD_INSNL(ret, nd_line(node), jump,   end_label);
    
    ADD_LABEL(ret, else_label);
    ADD_SEQ  (ret, else_seq);
    
    ADD_LABEL(ret, end_label);
    
    break;
  }
  case NODE_CASE:{
    NODE* vals;
    NODE* val;
    NODE* tempnode = node;
    LABEL* endlabel, *elselabel;
    DECL_ANCHOR(head);
    DECL_ANCHOR(body_seq);
    DECL_ANCHOR(cond_seq);
    VALUE special_literals = rb_ary_new();
    
    COMPILE(head, "case base", node->nd_head);
    
    node = node->nd_body;
    type = nd_type(node);

    if(type != NODE_WHEN){
      COMPILE_ERROR(("NODE_CASE: unexpected node. must be NODE_WHEN, byt %s", node_name(type)));
    }
    
    endlabel = NEW_LABEL(nd_line(node));
    elselabel = NEW_LABEL(nd_line(node));
    
    ADD_SEQ (ret, head); /* case VAL */
    
    while(type == NODE_WHEN){
      LABEL *l1;
      
      l1  = NEW_LABEL(nd_line(node));
      ADD_LABEL(body_seq, l1);
      ADD_INSN (body_seq, nd_line(node), pop);
      COMPILE_ (body_seq, "when body", node->nd_body, poped);
      ADD_INSNL(body_seq, nd_line(node), jump, endlabel);

      vals = node->nd_head;
      if(vals && nd_type(vals) == NODE_ARRAY){
        while(vals){
          VALUE lit;
          
          val = vals->nd_head;

          if((lit = case_when_optimizable_literal(val)) && special_literals){
            rb_ary_push(special_literals, rb_ary_new3(2, lit, (VALUE)(l1) | 1));
          }
          else{
            special_literals = Qfalse;
          }

          COMPILE  (cond_seq, "when cond", val);
          ADD_INSN1(cond_seq, nd_line(node), topn, INT2FIX(1));
          ADD_SEND (cond_seq, nd_line(node), ID2SYM(idEqq), INT2FIX(1));
          ADD_INSNL(cond_seq, nd_line(node), if, l1);
          vals = vals->nd_next;
        }
      }
      else{
        COMPILE_ERROR(("NODE_CASAE: must be NODE_ARRAY, but %s\n", node_name(nd_type(vals))));
      }
      
      node = node->nd_next;
      if(!node){
        break;
      }
      type = nd_type(node);
    }
    /* else */
    if(node){
      ADD_LABEL(cond_seq, elselabel);
      ADD_INSN (cond_seq, nd_line(node), pop);
      COMPILE_ (cond_seq, "else", node, poped);
      ADD_INSNL(cond_seq, nd_line(node), jump, endlabel);
    }
    else{
      debugs("== else(implicit)\n");
      ADD_LABEL(cond_seq, elselabel);
      ADD_INSN (cond_seq, nd_line(tempnode), pop);
      if(!poped){
        ADD_INSN(cond_seq, nd_line(tempnode), putnil);
      }
      ADD_INSNL(cond_seq, nd_line(tempnode), jump, endlabel);
    }
    
    if(special_literals){
      special_literals = (VALUE)NEW_WHILE(special_literals, 0, 0);
      ADD_INSN (ret, nd_line(tempnode), dup);
      ADD_INSN2(ret, nd_line(tempnode), opt_case_dispatch,
                special_literals, elselabel);
      rb_ary_push(iseqobj->compile_data->mark_ary, special_literals);
    }

    ADD_SEQ  (ret, cond_seq);
    ADD_SEQ  (ret, body_seq);
    ADD_LABEL(ret, endlabel);
    break;
  }
  case NODE_WHEN:{
    NODE* vals;
    NODE* val;
    LABEL* endlabel;
    DECL_ANCHOR(body_seq);

    endlabel = NEW_LABEL(nd_line(node));

    while(nd_type(node) == NODE_WHEN){
      LABEL *l1  = NEW_LABEL(nd_line(node));
      ADD_LABEL(body_seq, l1);
      COMPILE_ (body_seq, "when", node->nd_body, poped);
      ADD_INSNL(body_seq, nd_line(node), jump, endlabel);
      
      vals = node->nd_head;
      if(vals && nd_type(vals) == NODE_ARRAY){
        while(vals){
          val = vals->nd_head;
          COMPILE  (ret, "when2", val);
          ADD_INSNL(ret, nd_line(val), if, l1);
          vals = vals->nd_next;
        }
      }
      else{
        COMPILE_ERROR(("NODE_WHEN: must be NODE_ARRAY, but %s\n", node_name(nd_type(vals))));
      }
      node = node->nd_next;
    }
    /* else */
    COMPILE_ (ret, "else", node, poped);
    ADD_INSNL(ret, nd_line(node), jump, endlabel);

    ADD_SEQ  (ret, body_seq);
    ADD_LABEL(ret, endlabel);
    
    break;
  }
  case NODE_OPT_N:{
    /* unsupport currently */
    COMPILE_ERROR(("BUG: unknown node: NODE_OPT_N"));
    break;
  }
  case NODE_WHILE:
  case NODE_UNTIL:{
    LABEL *prev_start_label   = iseqobj->compile_data->start_label;
    LABEL *prev_end_label     = iseqobj->compile_data->end_label;
    LABEL *prev_redo_label    = iseqobj->compile_data->redo_label;
    VALUE prev_loopval_popped = iseqobj->compile_data->loopval_popped;
    struct iseq_compile_data_ensure_node_stack *enlp = iseqobj->compile_data->ensure_node_stack;

    LABEL * next_label  = iseqobj->compile_data->start_label =
      NEW_LABEL(nd_line(node)); /* next  */
    LABEL * redo_label  = iseqobj->compile_data->redo_label  =
      NEW_LABEL(nd_line(node)); /* redo  */
    LABEL * break_label = iseqobj->compile_data->end_label   =
      NEW_LABEL(nd_line(node)); /* break */
    LABEL * end_label = NEW_LABEL(nd_line(node));

    iseqobj->compile_data->loopval_popped = poped;
    iseqobj->compile_data->ensure_node_stack = 0;

    ADD_INSNL(ret, nd_line(node), jump, next_label);

    ADD_LABEL(ret, redo_label);
    COMPILE_ (ret, "while body", node->nd_body, 1);

    ADD_LABEL(ret, next_label); /* next */
    if(type == NODE_WHILE){
      compile_branch_condition(self, iseqobj, ret,
                               node->nd_cond, redo_label, end_label);
    }
    else{
      /* untile */
      compile_branch_condition(self, iseqobj, ret,
                               node->nd_cond, end_label, redo_label);
    }

    ADD_LABEL(ret, end_label);
    if(!poped){
      ADD_INSN(ret, nd_line(node), putnil);
    }
    ADD_LABEL(ret, break_label); /* braek */
    
    ADD_CATCH_ENTRY(CATCH_TYPE_BREAK, redo_label, break_label, 0, 0, break_label);
    ADD_CATCH_ENTRY(CATCH_TYPE_NEXT, redo_label, break_label,  0, 0, next_label);
    ADD_CATCH_ENTRY(CATCH_TYPE_REDO, redo_label, break_label,  0, 0, redo_label);
    
    
    iseqobj->compile_data->start_label = prev_start_label;
    iseqobj->compile_data->end_label   = prev_end_label;
    iseqobj->compile_data->redo_label  = prev_redo_label;
    iseqobj->compile_data->loopval_popped = prev_loopval_popped;
    iseqobj->compile_data->ensure_node_stack = enlp;
    break;
  }
  case NODE_ITER:
  case NODE_FOR:{
    VALUE prevblock = iseqobj->compile_data->current_block;
    LABEL * retry_label = NEW_LABEL(nd_line(node));
    LABEL * retry_end_l = NEW_LABEL(nd_line(node));
    VALUE name = rb_str_new2("b@");
    rb_str_concat(name, iseqobj->name);

    ADD_LABEL(ret, retry_label);
    if(nd_type(node) == NODE_FOR){
      COMPILE(ret, "iter caller (for)", node->nd_iter);
      
      iseqobj->compile_data->current_block =
        NEW_CHILD_ISEQOBJ(node, make_name_for_block(iseqobj),
                          self, ISEQ_TYPE_BLOCK);
      
      ADD_SEND_R(ret, nd_line(node), ID2SYM(idEach), INT2FIX(0),
                 iseqobj->compile_data->current_block, INT2FIX(0));
      if(poped){
        ADD_INSN(ret, nd_line(node), pop);
      }
    }
    else{
      iseqobj->compile_data->current_block =
        NEW_CHILD_ISEQOBJ(node, make_name_for_block(iseqobj),
                          self, ISEQ_TYPE_BLOCK);
      COMPILE_(ret, "iter caller", node->nd_iter, poped);
    }
    ADD_LABEL(ret, retry_end_l);
    rb_ary_push(iseqobj->iseq_mark_ary, iseqobj->compile_data->current_block);
    iseqobj->compile_data->current_block = prevblock;

    ADD_CATCH_ENTRY(CATCH_TYPE_RETRY, retry_label, retry_end_l, 0, 0, retry_label);
    break;
  }
  case NODE_BREAK:{
    unsigned long level = 0;
    
    if(iseqobj->compile_data->redo_label != 0){
      /* while/until */
      COMPILE_(ret, "break val(while/until)", node->nd_stts,
               iseqobj->compile_data->loopval_popped);
      add_ensure_iseq(ret, iseqobj, self);
      ADD_INSNL(ret, nd_line(node), jump, iseqobj->compile_data->end_label);
    }
    else if(iseqobj->type == ISEQ_TYPE_BLOCK){
    break_by_jump:
      /* escape from block */
      COMPILE(ret, "break val(block)", node->nd_stts);
      ADD_INSN1(ret, nd_line(node), throw, I2F(level | 0x02) /* TAG_BREAK */);
    }
    else{
      yarv_iseq_t *ip = iseqobj->parent_iseqobj;
      
      while(ip){
        level++;
        if(ip->type == ISEQ_TYPE_BLOCK){
          level <<= 16;
          goto break_by_jump;
        }
        else if(ip->compile_data->redo_label != 0){
          level = 0x8000;
          if(ip->compile_data->loopval_popped == 0){
            level |= 0x4000;
          }
          goto break_by_jump;
        }
        ip = iseqobj->parent_iseqobj;
      }
      COMPILE_ERROR(("can't put break"));
    }

    break;
  }
  case NODE_NEXT:{
    unsigned long level = 0;
    
    if(iseqobj->compile_data->redo_label != 0){
      add_ensure_iseq(ret, iseqobj, self);
      ADD_INSNL(ret, nd_line(node), jump, iseqobj->compile_data->start_label);
    }
    else if(iseqobj->compile_data->end_label){
      COMPILE(ret, "next val", node->nd_stts);
      add_ensure_iseq(ret, iseqobj, self);
      ADD_INSNL(ret, nd_line(node), jump, iseqobj->compile_data->end_label);
    }
    else{
      yarv_iseq_t *ip = iseqobj->parent_iseqobj;
      while(ip){
        level = 0x8000;
        if(ip->type == ISEQ_TYPE_BLOCK){
          break;
        }
        else if(ip->compile_data->redo_label != 0){
          if(ip->compile_data->loopval_popped == 0){
            level |= 0x4000;
          }
          break;
        }
        ip = iseqobj->parent_iseqobj;
      }
      if(ip != 0){
        COMPILE(ret, "next val", node->nd_stts);
        add_ensure_iseq(ret, iseqobj, self);
        ADD_INSN1(ret, nd_line(node), throw, I2F(level | 0x03) /* TAG_NEXT */);
      }
      else{
        COMPILE_ERROR(("can't next"));
      }
    }
    break;
  }
  case NODE_REDO:{
    if(iseqobj->compile_data->redo_label){
      add_ensure_iseq(ret, iseqobj, self);
      ADD_INSNL(ret, nd_line(node), jump, iseqobj->compile_data->redo_label);
    }
    else if(iseqobj->compile_data->start_label){
      ADD_INSNL(ret, nd_line(node), jump, iseqobj->compile_data->start_label);
    }
    else{
      yarv_iseq_t *ip = iseqobj->parent_iseqobj;
      unsigned long level = 0x8000 | 0x4000;
      while(ip){
        if(ip->type == ISEQ_TYPE_BLOCK){
          break;
        }
        else if(ip->compile_data->redo_label != 0){
          break;
        }
        ip = iseqobj->parent_iseqobj;
      }
      if(ip != 0){
        add_ensure_iseq(ret, iseqobj, self);
        ADD_INSN1(ret, nd_line(node), throw, I2F(level | 0x05) /* TAG_REDO */);
      }
      else{
        COMPILE_ERROR(("can't redo"));
      }
    }
    break;
  }
  case NODE_RETRY:{
    if(iseqobj->type == ISEQ_TYPE_BLOCK ||
       iseqobj->type == ISEQ_TYPE_RESCUE){
      ADD_INSN (ret, nd_line(node), putnil);
      ADD_INSN1(ret, nd_line(node), throw, I2F(0x04) /* TAG_RETRY */);
    }
    else{
      COMPILE_ERROR(("can't put retry"));
    }
    break;
  }
  case NODE_BEGIN:{
    COMPILE_(ret, "NODE_BEGIN", node->nd_body, poped);
    break;
  }
  case NODE_RESCUE:{
    LABEL *lstart = NEW_LABEL(nd_line(node));
    LABEL *lend   = NEW_LABEL(nd_line(node));
    LABEL *lcont  = NEW_LABEL(nd_line(node));
    VALUE rescue  = NEW_CHILD_ISEQOBJ(node->nd_resq, rb_str_new2("rescue"),
                                      self, ISEQ_TYPE_RESCUE);
    
    ADD_LABEL(ret, lstart);
    COMPILE(ret, "rescue head", node->nd_head);
    ADD_LABEL(ret, lend);
    if(node->nd_else){
      COMPILE(ret, "rescue else", node->nd_else);
    }
    ADD_INSN (ret, nd_line(node), nop);
    ADD_LABEL(ret, lcont);

    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    
    /* resgister catch entry */
    ADD_CATCH_ENTRY(CATCH_TYPE_RESCUE, lstart, lend, rescue, 0, lcont);
    ADD_CATCH_ENTRY(CATCH_TYPE_RETRY,  lend, lcont, 0, 0, lstart);
    break;
  }
  case NODE_RESBODY:{
    NODE *resq = node;
    NODE *narg;
        LABEL * label_miss, * label_hit;
    
    while(resq){
      label_miss = NEW_LABEL(nd_line(node));
      label_hit  = NEW_LABEL(nd_line(node));
      
      narg = resq->nd_args;
      while(narg){
        COMPILE  (ret, "rescue arg", narg->nd_head);
        ADD_INSN2(ret, nd_line(node), getdynamic, I2F(1), I2F(0));
        ADD_SEND (ret, nd_line(node), ID2SYM(idEqq), INT2FIX(1));
        ADD_INSNL(ret, nd_line(node), if, label_hit);
        narg = narg->nd_next;
      }
      if(resq->nd_args == 0){
        ADD_INSN1(ret, nd_line(node), putobject, rb_eStandardError);
        ADD_INSN2(ret, nd_line(node), getdynamic, I2F(1), I2F(0));
        ADD_SEND (ret, nd_line(node), ID2SYM(idEqq), INT2FIX(1));
        ADD_INSNL(ret, nd_line(node), if, label_hit);
      }
      ADD_INSNL(ret, nd_line(node), jump, label_miss);
      ADD_LABEL(ret, label_hit);
      COMPILE  (ret, "resbody body", resq->nd_body);
      ADD_INSN (ret, nd_line(node), end);
      ADD_LABEL(ret, label_miss);
      resq = resq->nd_head;
    }
    break;
  }
  case NODE_ENSURE:{
    DECL_ANCHOR(ensr);
    VALUE ensure = NEW_CHILD_ISEQOBJ(node->nd_ensr, rb_str_new2("ensure"),
                                     self, ISEQ_TYPE_ENSURE);
    LABEL * lstart = NEW_LABEL(nd_line(node));
    LABEL * lend   = NEW_LABEL(nd_line(node));
    LABEL * lcont  = NEW_LABEL(nd_line(node));
    VALUE prev_in_ensure = iseqobj->compile_data->in_ensure;
    struct iseq_compile_data_ensure_node_stack enl = {
      node->nd_ensr,
      iseqobj->compile_data->ensure_node_stack /* prev */
    };


    iseqobj->compile_data->in_ensure = Qtrue;
    COMPILE_POPED(ensr, "ensure ensr", node->nd_ensr);

    iseqobj->compile_data->in_ensure = prev_in_ensure;
    iseqobj->compile_data->ensure_node_stack = &enl;
    
    ADD_LABEL(ret, lstart);
    COMPILE_(ret, "ensure head", node->nd_head, poped);
    ADD_LABEL(ret, lend);
    if(ensr->anchor.next == 0){
      ADD_INSN(ret, nd_line(node), nop);
    }
    else{
      ADD_SEQ(ret, ensr);
    }
    ADD_LABEL(ret, lcont);

    ADD_CATCH_ENTRY(CATCH_TYPE_ENSURE, lstart, lend, ensure, 0, lcont);
    iseqobj->compile_data->ensure_node_stack = enl.prev;
    break;
  }
    
  case NODE_AND:
  case NODE_OR:{
    LABEL * end_label = NEW_LABEL(nd_line(node));
    COMPILE(ret, "nd_1st", node->nd_1st);
    if(!poped){
      ADD_INSN(ret, nd_line(node), dup);
    }
    
    if(type == NODE_AND){
      ADD_INSNL(ret, nd_line(node), unless, end_label);
    }
    else{
      ADD_INSNL(ret, nd_line(node), if, end_label);
    }
    ADD_INSN (ret, nd_line(node), pop);
    COMPILE_ (ret, "nd_2nd", node->nd_2nd, poped);
    ADD_LABEL(ret, end_label);
    
    break;
  }
    
  case NODE_NOT:{
    COMPILE (ret, "value", node->nd_body);
    ADD_INSN(ret, nd_line(node), putnot);
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
    
  case NODE_MASGN:{
    NODE *rhsn = node->nd_value;
    NODE *lhsn = node->nd_head;
    DECL_ANCHOR(mlhs_seq);
    
    int rlen, llen, i, max, si = 0;
    VALUE lhs_splat = Qfalse;

    llen = 0;
    while(lhsn){
      make_masgn_lhs(self, iseqobj, mlhs_seq, lhsn->nd_head);
      llen += 1;
      lhsn  = lhsn->nd_next;
    }
    
    if(node->nd_args && (long)node->nd_args != -1){
      make_masgn_lhs(self, iseqobj, mlhs_seq, node->nd_args);
      lhs_splat = Qtrue;
    }
    
    
    switch(nd_type(rhsn)){
    case NODE_ARRAY:
      rlen = rhsn->nd_alen;
      max = rlen > llen ? rlen : llen;
      debugs("l: %d, r: %d\n", llen, rlen);
      for(i=0; i<max; i++){
        if(i<rlen && i<llen){
          /* a, b = c, d*/
          COMPILE(ret, "masgn val1", rhsn->nd_head);
          rhsn = rhsn->nd_next;
        }
        else if(i<rlen){
          if(lhs_splat){
            while(rhsn){
              si++;
              COMPILE(ret, "masgn rhs for lhs splat", rhsn->nd_head);
              rhsn = rhsn->nd_next;
            }
            break;
          }
          else{
            /* a, b = c, d, e */
            COMPILE_POPED(ret, "masgn val2", rhsn->nd_head);
            rhsn = rhsn->nd_next;
          }
        }
        else if(i<llen){
          /* a, b, c = c, d */
          ADD_INSN(ret, nd_line(node), putnil);
        }
      }
      
      if(lhs_splat){
        ADD_INSN1(ret, nd_line(node), newarray, I2F(si));
      }
      break;
    case NODE_TO_ARY:
      COMPILE  (ret, "rhs to ary", rhsn->nd_head);
      ADD_INSN2(ret, nd_line(node), expandarray, I2F(llen), lhs_splat);
      break;
      
    case NODE_SPLAT:
      COMPILE  (ret, "rhs to ary (splat)", rhsn->nd_head);
      ADD_INSN2(ret, nd_line(node), expandarray, I2F(llen), lhs_splat);
      break;

    case NODE_ARGSCAT:{
      NODE *ary = rhsn->nd_head;
      int idx = 0;
      
      while(ary){
        if(idx < llen || lhs_splat){
          COMPILE(ret, "rhs aggscat each head", ary->nd_head);
        }
        else{
          COMPILE_POPED(ret, "rhs aggscat each head (popped)", ary->nd_head);
        }
        ary = ary->nd_next;
        idx++;
      }
      rlen = rhsn->nd_head->nd_alen;
      if(llen > idx){
        COMPILE  (ret, "rhs to ary (argscat/splat)", rhsn->nd_body);
        ADD_INSN2(ret, nd_line(node), expandarray, I2F(llen - idx), lhs_splat);
      }
      else if(lhs_splat){
        COMPILE  (ret, "rhs to ary (argscat/splat)", rhsn->nd_body);
        ADD_INSN2(ret, nd_line(node), expandarray, I2F(llen - idx), lhs_splat);
      }
      break;
    }
    default:
      rb_bug("unknown rhs: %s", node_name(nd_type(node->nd_value)));
    }
    ADD_SEQ(ret, REVERSE_LIST(mlhs_seq));

    if(!poped){
      ADD_INSN(ret, nd_line(node), putnil);
    }
    break;
  }
  case NODE_LASGN:{
    int idx = get_root_iseq_localsize(iseqobj) + 2 - node->nd_cnt;
    
    debugs("lvar: %d\n", idx);
    COMPILE(ret, "lvalue", node->nd_value);
    
    if(!poped){
      ADD_INSN(ret, nd_line(node), dup);
    }
    ADD_INSN1(ret, nd_line(node), setlocal, I2F(idx));
    
    break;
  }
  case NODE_DASGN:
  case NODE_DASGN_CURR:{
    int idx, lv, ls;
    COMPILE(ret, "dvalue", node->nd_value);
    
    debugp_param("dassn id", rb_str_new2(rb_id2name(node->nd_vid)));

    if(!poped){
      ADD_INSN(ret, nd_line(node), dup);
    }
    if(nd_type(node) == NODE_DASGN){
      idx = get_dyna_var_idx(self, node->nd_vid, &lv, &ls);
    }
    else{
      idx = get_dyna_var_idx_at(self, node->nd_vid);
      ls = iseqobj->local_size;
      lv = 0;
    }
    if(idx < 0){
      debugi("unknown id", node->nd_vid);
      COMPILE_ERROR(("NODE_DASGN error"));
    }
    ADD_INSN2(ret, nd_line(node), setdynamic,
              I2F(ls - idx), I2F(lv));
    break;
  }
  case NODE_GASGN:{
    COMPILE(ret, "lvalue", node->nd_value);

    if(!poped){
      ADD_INSN(ret, nd_line(node), dup);
    }
    ADD_INSN1(ret, nd_line(node), setglobal, (((long)node->nd_entry) | 1));
    break;
  }
  case NODE_IASGN:{
    COMPILE(ret, "lvalue", node->nd_value);
    if(!poped){
      ADD_INSN(ret, nd_line(node), dup);
    }
    ADD_INSN1(ret, nd_line(node), setinstancevariable, ID2SYM(node->nd_vid));
    break;
  }
  case NODE_CDECL:{
    COMPILE(ret, "lvalue", node->nd_value);

    if(!poped){
      ADD_INSN(ret, nd_line(node), dup);
    }

    if(node->nd_vid){
      ADD_INSN (ret, nd_line(node), putnil);
      ADD_INSN1(ret, nd_line(node), setconstant, ID2SYM(node->nd_vid));
    }
    else{
      COMPILE  (ret, "nd_else->nd_head", node->nd_else->nd_head);
      ADD_INSN1(ret, nd_line(node), setconstant, ID2SYM(node->nd_else->nd_mid));
    }
    break;
  }
  case NODE_CVASGN:
  case NODE_CVDECL:{
    COMPILE(ret, "cvasgn val", node->nd_value);
    if(!poped){
      ADD_INSN(ret, nd_line(node), dup);
    }
    ADD_INSN2(ret, nd_line(node), setclassvariable, ID2SYM(node->nd_vid),
              nd_type(node) == NODE_CVDECL ? Qtrue : Qfalse);
    break;
  }
  case NODE_OP_ASGN1:{
    DECL_ANCHOR(args);
    int   argc;
    ID    id = node->nd_mid;

    /*
     * a[x] (op)= y
     *
     * eval a    # a
     * eval x    # a x
     * dupn 2    # a x a x
     * send :[]  # a x a[x]
     * eval y    # a x a[x] y
     * send op   # a x a[x]+y
     * send []=  # ret
     */
    COMPILE(ret, "NODE_OP_ASGN1 recv", node->nd_recv);

    compile_array(self, iseqobj, args, node->nd_args->nd_next, Qfalse);
    POP_ELEMENT(args); POP_ELEMENT(args);
    argc = node->nd_args->nd_alen - 2;

    ADD_SEQ  (ret, args);
    ADD_INSN1(ret, nd_line(node), dupn, I2F(argc+1));
    ADD_SEND (ret, nd_line(node), ID2SYM(idAREF), I2F(argc));

    if(id == 0 || id == 1){
      /* 0: or, 1: and
        a[x] ||= y

        unless/if a[x]
          a[x]= y
        else
          nil
        end
       */
      LABEL * label = NEW_LABEL(nd_line(node));
      LABEL * lfin  = NEW_LABEL(nd_line(node));
      
      if(id == 0){
        ADD_INSN (ret, nd_line(node), dup);
        ADD_INSNL(ret, nd_line(node), if, label);
        ADD_INSN (ret, nd_line(node), pop);
      }
      else{
        ADD_INSNL(ret, nd_line(node), unless, label);
      }
      COMPILE  (ret, "NODE_OP_ASGN1 args->head: ", node->nd_args->nd_head);
      ADD_SEND (ret, nd_line(node), ID2SYM(idASET), I2F(argc+1));
      ADD_INSNL(ret, nd_line(node), jump, lfin);
      ADD_LABEL(ret, label);
      if(id == 1){ /* and */
        ADD_INSN (ret, nd_line(node), putnil);
      }
      ADD_LABEL(ret, lfin);
    }
    else{
      COMPILE (ret, "NODE_OP_ASGN1 args->head: ", node->nd_args->nd_head);
      ADD_SEND(ret, nd_line(node), ID2SYM(id), I2F(1));
      ADD_SEND(ret, nd_line(node), ID2SYM(idASET), I2F(argc+1));
    }

    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }

    break;
  }
  case NODE_OP_ASGN2:{
    ID    atype = node->nd_next->nd_mid;
    LABEL * lfin  = NEW_LABEL(nd_line(node));
    LABEL * lcfin = NEW_LABEL(nd_line(node));
    /*
      class C; attr_accessor :c; end
      r = C.new
      r.a &&= v # asgn2

      eval r    # r
      dup       # r r
      eval r.a  # r o
      
      # or
      dup       # r o o
      if lcfin  # r o
      pop       # r
      eval v    # r v
      send a=   # v
      jump lfin # v
      
    lcfin:      # r o
      swap      # o r
      pop       # o

    lfin:       # v

      # and
      dup       # r o o
      unless lcfin
      pop       # r
      eval v    # r v
      send a=   # v
      jump lfin # v

      # others
      eval v    # r o v
      send ??   # r w
      send a=   # w
      
     */

    COMPILE  (ret, "NODE_OP_ASGN2#recv", node->nd_recv);
    ADD_INSN (ret, nd_line(node), dup);
    ADD_SEND(ret, nd_line(node), ID2SYM(node->nd_next->nd_vid), I2F(0));

    if(atype == 0 || atype == 1){ /* 0: OR or 1: AND */
      ADD_INSN (ret, nd_line(node), dup);
      if(atype == 0){
        ADD_INSNL(ret, nd_line(node), if, lcfin);
      }
      else{
        ADD_INSNL(ret, nd_line(node), unless, lcfin);
      }
      ADD_INSN (ret, nd_line(node), pop);
      COMPILE  (ret, "NODE_OP_ASGN2 val", node->nd_value);
      ADD_SEND (ret, nd_line(node), ID2SYM(node->nd_next->nd_aid), I2F(1));
      ADD_INSNL(ret, nd_line(node), jump, lfin);
      
      ADD_LABEL(ret, lcfin);
      ADD_INSN (ret, nd_line(node), swap);
      ADD_INSN (ret, nd_line(node), pop);
      
      ADD_LABEL(ret, lfin);
    }
    else{
      COMPILE (ret, "NODE_OP_ASGN2 val", node->nd_value);
      ADD_SEND(ret, nd_line(node), ID2SYM(node->nd_next->nd_mid), I2F(1));
      ADD_SEND(ret, nd_line(node), ID2SYM(node->nd_next->nd_aid), I2F(1));
    }
    
    if(poped){
      /* we can apply more optimize */
      ADD_INSN (ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_OP_ASGN_AND:
  case NODE_OP_ASGN_OR:{
    LABEL * lfin = NEW_LABEL(nd_line(node));
    
    COMPILE (ret, "NODE_OP_ASGN_AND#nd_head", node->nd_head);
    ADD_INSN(ret, nd_line(node), dup);
    if(nd_type(node) == NODE_OP_ASGN_AND){
      ADD_INSNL(ret, nd_line(node), unless, lfin);
    }
    else{
      ADD_INSNL(ret, nd_line(node), if, lfin);
    }
    ADD_INSN (ret, nd_line(node), pop);
    COMPILE  (ret, "NODE_OP_ASGN_AND#nd_value", node->nd_value);
    ADD_LABEL(ret, lfin);
    
    if(poped){
      /* we can apply more optimize */
      ADD_INSN (ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_CALL:
  case NODE_FCALL:
  case NODE_VCALL:{ /* VCALL: variable or call */
    /*
      call:  obj.method(...)
      fcall: func(...)
      vcall: func()
     */
    DECL_ANCHOR(recv);
    DECL_ANCHOR(args);
    VALUE argc;
    VALUE flag  = 0;
    VALUE block = 0;
    VALUE parent_block = iseqobj->compile_data->current_block;
    iseqobj->compile_data->current_block = Qfalse;
    
#if SUPPORT_GOTO
    /* only joke */
    {
      static ID goto_id;
      static ID label_id;
      VALUE label;
      VALUE label_sym;

      if(goto_id == 0){
        goto_id  = rb_intern("__goto__");
        label_id = rb_intern("__label__");
      }
      
      if(nd_type(node) == NODE_FCALL &&
         (node->nd_mid == goto_id || node->nd_mid == label_id)){
        if(nd_type(node->nd_args->nd_head) == NODE_LIT &&
           SYMBOL_P(node->nd_args->nd_head->nd_lit)){

          label_sym = label = node->nd_args->nd_head->nd_lit;
          if((label = rb_hash_aref(iseqobj->compile_data, label_sym)) == Qnil){
            rb_hash_aset(iseqobj->compile_data, label_sym,
                         label = NEW_LABEL(nd_line(node)));
          }
        }
        else{
          rb_bug("illegal goto/label format");
        }
        
        
        if(node->nd_mid == goto_id){
          ADD_INSNL(ret, nd_line(node), jump, label);
        }
        else{
          ADD_LABEL(ret, label);
        }
        break;
      }
    }
#endif

    /* reciever */
    if(type == NODE_CALL){
      COMPILE(recv, "recv", node->nd_recv);
    }
    else if(type == NODE_FCALL ||
            type == NODE_VCALL){
      ADD_INSN(recv, nd_line(node), putself);
    }

    /* args */
    if(type != NODE_VCALL && node->nd_args){
      if(nd_type(node->nd_args) == NODE_SPLAT){
        COMPILE(args, "args(splat)", node->nd_args->nd_head);
        argc  = I2F(1);
        flag |= VM_CALL_ARGS_SPLAT_BIT;
      }
      else if(nd_type(node->nd_args) == NODE_ARGSCAT){
        compile_array(self, iseqobj, args, node->nd_args->nd_head, Qfalse);
        POP_ELEMENT(args);
        
        argc = I2F(LIST_SIZE(args) + 1);
        COMPILE(args, "args(cat: splat)", node->nd_args->nd_body);

        flag |= VM_CALL_ARGS_SPLAT_BIT;
      }
      else{
        compile_array(self, iseqobj, args, node->nd_args, Qfalse);
        argc = OPERAND_AT(POP_ELEMENT(args), 0);
     }
    }
    else{
      argc = I2F(0);
    }

    ADD_SEQ(ret, recv);
    ADD_SEQ(ret, args);
    
    if(parent_block){
      if(parent_block & 1){
        flag |= VM_CALL_ARGS_BLOCKARG_BIT;
        ADD_SEQ(ret, (LINK_ANCHOR *)(parent_block & (~1)));
      }
      else{
        block = parent_block;
      }
    }

    debugp_param("call args argc", argc);
    debugp_param("call method",    ID2SYM(node->nd_mid));
    
    ADD_SEND_R(ret, nd_line(node), ID2SYM(node->nd_mid), argc,
               block, I2F(flag));
    
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_SUPER:{
    DECL_ANCHOR(args);
    VALUE argc;
    VALUE flag = 0;
    
    /* args */
    if(type != NODE_VCALL && node->nd_args){
      if(nd_type(node->nd_args) == NODE_SPLAT){
        COMPILE(args, "args(splat)", node->nd_args->nd_head);
        argc  = I2F(1);
        flag |= VM_CALL_ARGS_SPLAT_BIT;
      }
      else if(nd_type(node->nd_args) == NODE_ARGSCAT){
        compile_array(self, iseqobj, args, node->nd_args->nd_head, Qfalse);
        POP_ELEMENT(args);
        
        argc = LIST_SIZE(args) + 1;
        COMPILE(args, "args(cat: splat)", node->nd_args->nd_body);

        flag |= VM_CALL_ARGS_SPLAT_BIT;
      }
      else{
        compile_array(self, iseqobj, args, node->nd_args, Qfalse);
        argc = OPERAND_AT(POP_ELEMENT(args), 0);
     }
    }
    else{
      argc = I2F(0);
    }

    ADD_INSN(ret, nd_line(node), putnil); /* dummy reciever */
    ADD_SEQ(ret, args);
    ADD_INSN2(ret, nd_line(node), super, argc, I2F(flag));

    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_ZSUPER:{
    ADD_INSN(ret, nd_line(node), zsuper);

    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_ARRAY:{
    compile_array(self, iseqobj, ret, node, Qtrue);
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_ZARRAY:{
    if(!poped){
      ADD_INSN1(ret, nd_line(node), newarray, I2F(0));
    }
    break;
  }
  case NODE_VALUES:{
    NODE *n = node;
    while(n){
      COMPILE(ret, "values item", n->nd_head);
      n = n->nd_next;
    }
    ADD_INSN1(ret, nd_line(node), newarray, I2F(node->nd_alen));
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_HASH:{
    DECL_ANCHOR(list);
    VALUE size = 0;
    int type = node->nd_head ? nd_type(node->nd_head) : NODE_ZARRAY;
    
    switch(type){
    case NODE_ARRAY:{
      compile_array(self, iseqobj, list, node->nd_head, Qfalse);
      size = OPERAND_AT(POP_ELEMENT(list), 0);
      ADD_SEQ  (ret, list);
      break;
    } 
    case NODE_ZARRAY:
      size = I2F(0);
      break;
      
    default:
      COMPILE_ERROR(("can't make hash with this node: %s", node_name(type)));
    }

    ADD_INSN1(ret, nd_line(node), newhash, size);

    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_RETURN:{
    yarv_iseq_t *is = iseqobj;

    while(is){
      if(is->type == ISEQ_TYPE_TOP ||
         is->type == ISEQ_TYPE_CLASS){
        COMPILE_ERROR(("can't put return"));
        break;
      }
      else if(is->type == ISEQ_TYPE_RESCUE ||
              is->type == ISEQ_TYPE_ENSURE){
        is = is->parent_iseqobj;
      }
      else{
        COMPILE(ret, "return nd_stts(return val)", node->nd_stts);

        /* TODO: compile_data->in_ensure needed? */
        if(is->type == ISEQ_TYPE_METHOD){// && is->compile_data->in_ensure == Qfalse){
          add_ensure_iseq(ret, iseqobj, self);
          ADD_INSN(ret, nd_line(node), end);
        }
        else if(is->type == ISEQ_TYPE_BLOCK){// || is->compile_data->in_ensure == Qtrue){
          add_ensure_iseq(ret, iseqobj, self);
          ADD_INSN1(ret, nd_line(node), throw, I2F(0x01) /* TAG_RETURN */);
        }
        else{
          COMPILE_ERROR(("illegal return"));
        }
        break;
      }
    }
    
    break;
  }
  case NODE_YIELD:{
    DECL_ANCHOR(args);
    int argc;
    
    if(iseqobj->type == ISEQ_TYPE_TOP ||
       iseqobj->type == ISEQ_TYPE_CLASS){
      COMPILE_ERROR(("can't put yield"));
    }
    
    if(node->nd_head){
      if(nd_type(node->nd_head) == NODE_ARRAY){
        compile_array(self, iseqobj, args, node->nd_head, Qfalse);
        POP_ELEMENT(args);
        argc = LIST_SIZE(args);
        debugs("argc: %d\n", argc);
      }
      else{
        COMPILE(args, "nd_head(1)", node->nd_head);
        argc = 1;
      }
    }
    else{
      argc = 0;
    }
    ADD_SEQ  (ret, args);
    ADD_INSN2(ret, nd_line(node), yield, I2F(argc), I2F(0));

    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_LVAR:{
    if(!poped){
      int idx = get_root_iseq_localsize(iseqobj) + 2 - node->nd_cnt;
      debugs("idx: %d\n", idx);
      ADD_INSN1(ret, nd_line(node), getlocal, I2F(idx));
    }
    break;
  }
  case NODE_DVAR:{
    int lv, idx, ls;
    debugi("nd_vid", node->nd_vid);
    if(!poped){
      idx = get_dyna_var_idx(self, node->nd_vid, &lv, &ls);
      if(idx < 0){
        COMPILE_ERROR(("unknown dvar"));
      }
      ADD_INSN2(ret, nd_line(node), getdynamic, I2F(ls - idx), I2F(lv));
    }
    break;
  }
  case NODE_GVAR:{
    ADD_INSN1(ret, nd_line(node), getglobal, (((long)node->nd_entry) | 1));
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_IVAR:{
    debugi("nd_vid", node->nd_vid);
    if(!poped){
      ADD_INSN1(ret, nd_line(node), getinstancevariable, ID2SYM(node->nd_vid));
    }
    break;
  }
  case NODE_CONST:{
    debugi("nd_vid", node->nd_vid);
    if(!poped){
      if(iseqobj->compile_data->cached_const == Qfalse){
        LABEL * lstart = NEW_LABEL(nd_line(node));
        LABEL * lend   = NEW_LABEL(nd_line(node));
        
        ADD_LABEL(ret, lstart);
        ADD_INSN2(ret, nd_line(node), getinlinecache, NEW_INLINE_CACHE_ENTRY(), lend);
        /* ADD_INSN (ret, nd_line(node), pop); */
        /* ADD_INSN (ret, nd_line(node), putnil); */ /* target class */
        ADD_INSN1(ret, nd_line(node), getconstant, ID2SYM(node->nd_vid));
        ADD_INSN1(ret, nd_line(node), setinlinecache, lstart);
        ADD_LABEL(ret, lend);
      }
      else{
        ADD_INSN (ret, nd_line(node), putnil); /* target class */
        ADD_INSN1(ret, nd_line(node), getconstant, ID2SYM(node->nd_vid));
      }
    }
    break;
  }
  case NODE_CVAR:{
    if(!poped){
      ADD_INSN1(ret, nd_line(node), getclassvariable, ID2SYM(node->nd_vid));
    }
    break;
  }
  case NODE_NTH_REF:{
    ADD_INSN2(ret, nd_line(node), getspecial, I2F(node->nd_cnt), I2F(node->nd_nth << 1));
    break;
  }
  case NODE_BACK_REF:{
    ADD_INSN2(ret, nd_line(node), getspecial, I2F(node->nd_cnt), I2F(0x01 | (node->nd_nth << 1)));
    break;
  }
  case NODE_MATCH:
  case NODE_MATCH2:
  case NODE_MATCH3:{
    DECL_ANCHOR(recv);
    DECL_ANCHOR(val);

    if(nd_type(node) == NODE_MATCH){
      ADD_INSN1(recv, nd_line(node), putobject, node->nd_lit);
      ADD_INSN2(val, nd_line(node), getspecial, I2F(0), I2F(0));
    }
    else{
      COMPILE(recv, "reciever", node->nd_recv);
      COMPILE(val, "value", node->nd_value);
    }
    

#ifdef OPT_BASIC_OPERATIONS
    /* TODO: detect by node */
    if(recv->last == recv->anchor.next &&
       INSN_OF(recv->last) == BIN(putobject)){
      ADD_SEQ  (ret, val);
      ADD_INSN1(ret, nd_line(node), opt_regexpmatch1, OPERAND_AT(recv->last, 0));
    }
    else{
      ADD_SEQ(ret, recv);
      ADD_SEQ(ret, val);
      ADD_INSN(ret, nd_line(node), opt_regexpmatch2);
    }
#else
    ADD_SEQ(ret, recv);
    ADD_SEQ(ret, val);
    ADD_SEND(ret, nd_line(node), ID2SYM(idEqTilde), INT2FIX(1));
#endif
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_LIT:{
    debugp_param("lit", node->nd_lit);
    if(!poped){
      ADD_INSN1(ret, nd_line(node), putobject, node->nd_lit);
    }
    break;
  }
  case NODE_STR:{
    debugp_param("nd_lit", node->nd_lit);
    if(!poped){
      ADD_INSN1(ret, nd_line(node), putstring, node->nd_lit);
    }
    break;
  }
  case NODE_DSTR:{
    compile_dstr(self, iseqobj, ret, node);
    
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_XSTR:{
    ADD_INSN (ret, nd_line(node), putself);
    ADD_INSN1(ret, nd_line(node), putobject, node->nd_lit);
    ADD_SEND (ret, nd_line(node), ID2SYM(idBackquote), INT2FIX(1));
    
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_DXSTR:{
    ADD_INSN(ret, nd_line(node), putself);
    compile_dstr(self, iseqobj, ret, node);
    ADD_SEND(ret, nd_line(node), ID2SYM(idBackquote), INT2FIX(1));

    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_EVSTR:{
    COMPILE(ret, "nd_body", node->nd_body);
    
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    else{
      ADD_INSN(ret, nd_line(node), tostring);
    }
    break;
  }
  case NODE_DREGX:{
    compile_dstr(self, iseqobj, ret, node);
    ADD_INSN1(ret, nd_line(node), toregexp, I2F(node->nd_cflag));
    
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_DREGX_ONCE:{
    /* fix me */
    compile_dstr(self, iseqobj, ret, node);
    ADD_INSN1(ret, nd_line(node), toregexp, I2F(node->nd_cflag));
    
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_ARGS:{
    /* OK */
    COMPILE_ERROR(("BUG: should not reach here: compile_each#NODE_ARGS"));
    break;
  }
  case NODE_ARGSCAT:{
    NODE *n = node->nd_head;
    int i = 0;
    
    while(n){
      i++;
      COMPILE(ret, "argscat head", n->nd_head);
      n = n->nd_next;
    }
    n = node->nd_body;
    while(n){
      i++;
      COMPILE(ret, "argscat body", n->nd_head);
      n = n->nd_next;
    }
    ADD_INSN1(ret, nd_line(node), newarray, I2F(i));
    break;
  }
  case NODE_ARGSPUSH:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_ARGSPUSH"));
    break;
  }
  case NODE_SPLAT:{
    COMPILE(ret, "splat", node->nd_head);
    break;
  }
  case NODE_TO_ARY:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_TO_ARY"));
    break;
  }
  case NODE_SVALUE:{
    COMPILE(ret, "svalue head", node->nd_head);
    break;
  }
  case NODE_BLOCK_ARG:{
    iseqobj->arg_block = node->nd_cnt - 2;
    iseqobj->arg_simple = 0;
    
    break;
  }
  case NODE_BLOCK_PASS:{
    VALUE prevblock   = iseqobj->compile_data->current_block;
    LABEL * retry_label = NEW_LABEL(nd_line(node));
    LABEL * retry_end_l = NEW_LABEL(nd_line(node));
    DECL_ANCHOR(current_block);
    
    iseqobj->compile_data->current_block = (VALUE)current_block | 1;
    COMPILE(current_block, "block pass proc", node->nd_body);
    
    ADD_LABEL(ret, retry_label);
    COMPILE_ (ret, "iter caller (NODE_BLOCK_PASS)", node->nd_iter, poped);
    ADD_LABEL(ret, retry_end_l);
    
    iseqobj->compile_data->current_block = prevblock;
    ADD_CATCH_ENTRY(CATCH_TYPE_RETRY, retry_label, retry_end_l, 0, 0, retry_label);
    break;
  }
  case NODE_DEFN:{
    VALUE iseq = NEW_ISEQOBJ(node->nd_defn,
                             rb_str_new2(rb_id2name(node->nd_mid)),
                             ISEQ_TYPE_METHOD);
    
    debugp_param("defn/iseq", iseq);

    ADD_INSN2(ret, nd_line(node), methoddef, ID2SYM(node->nd_mid), iseq);
    if(!poped){
      ADD_INSN(ret, nd_line(node), putnil);
    }
    debugp_param("defn", iseq);
    break;
  }
  case NODE_DEFS:{
    VALUE iseq = NEW_ISEQOBJ(node->nd_defn,
                             rb_str_new2(rb_id2name(node->nd_mid)),
                             ISEQ_TYPE_METHOD);

    debugp_param("defs/iseq", iseq);

    COMPILE(ret, "defs: recv", node->nd_recv);
    ADD_INSN2(ret, nd_line(node), singletonmethoddef, ID2SYM(node->nd_mid), iseq);
    if(!poped){
      ADD_INSN(ret, nd_line(node), putnil);
    }
    break;
  }
  case NODE_ALIAS:{
    VALUE s1, s2;
    
    if(nd_type(node->u1.node) != NODE_LIT ||
       nd_type(node->u2.node) != NODE_LIT){
      rb_bug("alias args must be NODE_LIT");
    }
    s1 = node->u1.node->nd_lit;
    s2 = node->u2.node->nd_lit;

    ADD_INSN3(ret, nd_line(node), alias, Qfalse, ID2SYM(rb_to_id(s1)), ID2SYM(rb_to_id(s2)));
    if(!poped){
      ADD_INSN(ret, nd_line(node), putnil);
    }
    break;
  }
  case NODE_VALIAS:{
    ADD_INSN3(ret, nd_line(node), alias, Qtrue, ID2SYM(node->u1.id), ID2SYM(node->u2.id));
    if(!poped){
      ADD_INSN(ret, nd_line(node), putnil);
    }
    break;
  }
  case NODE_UNDEF:{
    if(nd_type(node->u2.node) != NODE_LIT){
      rb_bug("undef args must be NODE_LIT");
    }
    ADD_INSN1(ret, nd_line(node), undef, ID2SYM(rb_to_id(node->u2.node->nd_lit)));
    if(!poped){
      ADD_INSN(ret, nd_line(node), putnil);
    }
    break;
  }
  case NODE_CLASS:{
    VALUE iseq =
      NEW_CHILD_ISEQOBJ(node->nd_body,
                        make_name_with_str("<class:%s>", rb_id2name(node->nd_cpath->nd_mid)),
                        self, ISEQ_TYPE_CLASS);
    
    COMPILE(ret, "cbase", node->nd_cpath->nd_head);
    COMPILE(ret, "super", node->nd_super);
    ADD_INSN2(ret, nd_line(node), classdef, ID2SYM(node->nd_cpath->nd_mid), iseq);
    ADD_INSN (ret, nd_line(node), popcref);
    
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_MODULE:{
    VALUE iseq =
      NEW_CHILD_ISEQOBJ(node->nd_body,
                        make_name_with_str("<module:%s>", rb_id2name(node->nd_cpath->nd_mid)),
                        self, ISEQ_TYPE_CLASS);
    
    COMPILE  (ret, "mbase", node->nd_cpath->nd_head);
    ADD_INSN2(ret, nd_line(node), moduledef, ID2SYM(node->nd_cpath->nd_mid), iseq);
    ADD_INSN (ret, nd_line(node), popcref);
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_SCLASS:{
    VALUE iseq = NEW_ISEQOBJ(node->nd_body, rb_str_new2("singletonclass"),
                             ISEQ_TYPE_CLASS);

    COMPILE (ret, "sclass#recv", node->nd_recv);
    ADD_INSN1(ret, nd_line(node), singletonclassdef, iseq);
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_COLON2:{
    
    if(rb_is_const_id(node->nd_mid)){
      /* constant */
      debugi("nd_mid", node->nd_mid);
      if(iseqobj->compile_data->cached_const == Qfalse){
        LABEL * lstart = NEW_LABEL(nd_line(node));
        LABEL * lend   = NEW_LABEL(nd_line(node));
        
        /* add cache insn */
        iseqobj->compile_data->cached_const = Qtrue;
        ADD_LABEL(ret, lstart);
        ADD_INSN2(ret, nd_line(node), getinlinecache, NEW_INLINE_CACHE_ENTRY(), lend);
        ADD_INSN (ret, nd_line(node), pop);
        COMPILE  (ret, "colon2#nd_head", node->nd_head);
        ADD_INSN1(ret, nd_line(node), getconstant, ID2SYM(node->nd_mid));
        ADD_INSN1(ret, nd_line(node), setinlinecache, lstart);
        ADD_LABEL(ret, lend);
        iseqobj->compile_data->cached_const = Qfalse;
      }
      else{
        COMPILE(ret, "colon2#nd_head", node->nd_head);
        ADD_INSN1(ret, nd_line(node), getconstant, ID2SYM(node->nd_mid));
      }
    }
    else{
      /* function call */
      ADD_INSN(ret, nd_line(node), putself);
      COMPILE (ret, "colon2#nd_head", node->nd_head);
      ADD_SEND(ret, nd_line(node), ID2SYM(node->nd_mid), INT2FIX(1));
    }
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_COLON3:{
    debugi("colon3#nd_mid", node->nd_mid);
    if(iseqobj->compile_data->cached_const == Qfalse){
      LABEL * lstart = NEW_LABEL(nd_line(node));
      LABEL * lend   = NEW_LABEL(nd_line(node));

      /* add cache insn */
      ADD_LABEL(ret, lstart);
      ADD_INSN2(ret, nd_line(node), getinlinecache, NEW_INLINE_CACHE_ENTRY(), lend);
      ADD_INSN (ret, nd_line(node), pop);
      ADD_INSN1(ret, nd_line(node), putobject     , rb_cObject);
      ADD_INSN1(ret, nd_line(node), getconstant   , ID2SYM(node->nd_mid));
      ADD_INSN1(ret, nd_line(node), setinlinecache, lstart);
      ADD_LABEL(ret, lend);
    }
    else{
      ADD_INSN1(ret, nd_line(node), putobject, rb_cObject);
      ADD_INSN1(ret, nd_line(node), getconstant, ID2SYM(node->nd_mid));
    }
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_CREF:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_CREF"));
    break;
  }
  case NODE_DOT2:
  case NODE_DOT3:{
    int flag = type == NODE_DOT2 ? INT2FIX(0) : INT2FIX(1);
    COMPILE(ret, "min", (NODE*)node->nd_beg);
    COMPILE(ret, "max", (NODE*)node->nd_end);
    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
      ADD_INSN(ret, nd_line(node), pop);
    }
    else{
      ADD_INSN1(ret, nd_line(node), newrange, flag);
    }
    break;
  }
  case NODE_FLIP2:
  case NODE_FLIP3:{
    LABEL * lend   = NEW_LABEL(nd_line(node));
    LABEL * lfin   = NEW_LABEL(nd_line(node));
    LABEL * ltrue  = NEW_LABEL(nd_line(node));
    
    ADD_INSN2(ret, nd_line(node), getspecial, I2F(node->nd_cnt), I2F(0));
    ADD_INSNL(ret, nd_line(node), if, lend);

    /* *flip == 0 */
    COMPILE  (ret, "flip2 beg", node->nd_beg);
    ADD_INSN (ret, nd_line(node), dup);
    ADD_INSNL(ret, nd_line(node), unless, lfin);
    if(nd_type(node) == NODE_FLIP3){
      ADD_INSN (ret, nd_line(node), dup);
      ADD_INSN2(ret, nd_line(node), setspecial, I2F(node->nd_cnt), I2F(0));
      ADD_INSNL(ret, nd_line(node), jump, lfin);
    }
    else{
      ADD_INSN2(ret, nd_line(node), setspecial, I2F(node->nd_cnt), I2F(0));
    }

    /* *flip == 1 */
    ADD_LABEL(ret, lend);
    COMPILE  (ret, "flip2 end", node->nd_end);
    ADD_INSNL(ret, nd_line(node), unless, ltrue);
    ADD_INSN1(ret, nd_line(node), putobject, Qfalse);
    ADD_INSN2(ret, nd_line(node), setspecial, I2F(node->nd_cnt), I2F(0));
    
    ADD_LABEL(ret, ltrue);
    ADD_INSN1(ret, nd_line(node), putobject, Qtrue);
    
    ADD_LABEL(ret, lfin);
    
    break;
  }
  case NODE_ATTRSET:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_ATTRSET"));
    break;
  }
  case NODE_SELF:{
    if(!poped){
      ADD_INSN(ret, nd_line(node), putself);
    }

    break;
  }
  case NODE_NIL:{
    if(!poped){
      ADD_INSN(ret, nd_line(node), putnil);
    }
    break;
  }
  case NODE_TRUE:{
    if(!poped){
      ADD_INSN1(ret, nd_line(node), putobject, Qtrue);
    }
    break;
  }
  case NODE_FALSE:{
    if(!poped){
      ADD_INSN1(ret, nd_line(node), putobject, Qfalse);
    }
    break;
  }
  case NODE_ERRINFO:{
    if(!poped){
      if(iseqobj->type == ISEQ_TYPE_RESCUE){
        ADD_INSN2(ret, nd_line(node), getdynamic, I2F(1), I2F(0));
      }
      else{
        ADD_INSN(ret, nd_line(node), putnil);
      }
    }
    break;
  }
  case NODE_DEFINED:{
    if(!poped){
      LABEL * lfinish = NEW_LABEL(nd_line(node));
      defined_expr(self, iseqobj, ret, node->nd_head, lfinish, Qtrue);
      ADD_LABEL(ret, lfinish);
    }
    break;
  }
  case NODE_POSTEXE:{
    COMPILE_ERROR(("BUG: unknown node: NODE_POSTEXE"));
    break;
  }
#ifdef C_ALLOCA
  case NODE_ALLOCA:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_ALLOCA"));
    break;
  }
#endif
  case NODE_BMETHOD:{
    /* block method, OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_BMETHOD"));
    break;
  }
  case NODE_MEMO:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_MEMO"));
    break;
  }
  case NODE_IFUNC:{
    /* OK */
    COMPILE_ERROR(("BUG: unknown node: NODE_IFUNC"));
    break;
  }
  case NODE_DSYM:{
    compile_dstr(self, iseqobj, ret, node);
    if(!poped){
      ADD_SEND(ret, nd_line(node), ID2SYM(idIntern), I2F(0));
    }
    else{
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  case NODE_ATTRASGN:{
    DECL_ANCHOR(recv);
    DECL_ANCHOR(args);
    VALUE argc;

    if(node->nd_args){
      compile_array(self, iseqobj, args, node->nd_args, Qfalse);
      argc = OPERAND_AT(POP_ELEMENT(args), 0);
    }
    else{
      argc = I2F(0); /* massign */
    }
    
    if(node->nd_recv == (NODE *)1){
      ADD_INSN(recv, nd_line(node), putself);
    }
    else{
      COMPILE(recv, "recv", node->nd_recv);
    }
    
    debugp_param("argc"   , argc);
    debugp_param("nd_mid",  ID2SYM(node->nd_mid));
    
    ADD_SEQ(ret, recv);
    ADD_SEQ(ret, args);

    ADD_SEND(ret, nd_line(node), ID2SYM(node->nd_mid), argc);

    if(poped){
      ADD_INSN(ret, nd_line(node), pop);
    }
    break;
  }
  default:
    COMPILE_ERROR(("BUG: unknown node: %s", node_name(type)));
    return Qnil;
  }

  debug_nodeprint_close();
  return COMPILE_OK;
}


/***************************/
/* instruction information */
/***************************/

#define TS_LONG     "L"
#define TS_OFFSET   "L"
#define TS_ULONG    "U"
#define TS_VALUE    "V"
#define TS_ID       "I"
#define TS_GENTRY   "G"
#define TS_IC       "M"
#define TS_CDHASH   "H"
#define TS_VARIABLE "."


#include "insns_info.inc"

/* some utility */

#if CPDEBUG > 5

#define INSN_CHECK(insn)                                         \
  if(insn < 0 ||                                                 \
     insn > sizeof(insn_name_info) / sizeof(insn_name_info[0])){ \
    rb_bug("(in)unknown instruction: %d", insn);                 \
  }                                                              \

#else

#define INSN_CHECK(insn) /* */

#endif

char *insn_name(int insn){
  INSN_CHECK(insn);
  return insn_name_info[insn];
}

int insn_len(int insn){
  INSN_CHECK(insn);
  return insn_len_info[insn];
}

char *insn_op_types(int insn){
  INSN_CHECK(insn);
  return insn_operand_info[insn];
}

int insn_op_type(int insn, int pos){
  int len = insn_len(insn) - 1;
  if(pos < len){
    return insn_operand_info[insn][pos];
  }
  else{
    return TS_LONG[0];
  }
}

static int insn_ret_num(int insn){
  INSN_CHECK(insn);
  return insn_stack_push_num_info[insn];
}

static int insn_data_length(INSN *insnobj){
  return insn_len(insnobj->insn_id);
}

static int insn_data_line_no(INSN *insnobj){
  return insn_len(insnobj->line_no);
}

VALUE insn_data_to_s_detail(INSN *insnobj){
  VALUE str = rb_str_new(0,0);
  char buff[0x100];

  snprintf(buff, sizeof(buff), "%-16s", insn_name(insnobj->insn_id));
  rb_str_cat2(str, buff);
  if(insnobj->operands){
    char *types = insn_op_types(insnobj->insn_id);
    int j;
    
    for(j=0;types[j];j++){
      char type = types[j];

      switch(type){
      case 'L':          /* label(destination position) */
        {
          char  buff[0x100];
          LABEL *lobj = (LABEL *)OPERAND_AT(insnobj, j);
          snprintf(buff, sizeof(buff), "<L%03d>", lobj->label_no);
          rb_str_concat(str, rb_str_new2(buff));
          break;
        }
        break;
      case 'U':          /* ulong */
      case 'V':          /* VALUE */
        rb_str_concat(str, rb_inspect(OPERAND_AT(insnobj, j)));
        break;
      case 'I':          /* ID */
        rb_str_concat(str, rb_inspect(OPERAND_AT(insnobj, j)));
        break;
      case 'G':
        {
          struct global_entry *entry = (struct global_entry*)
            (OPERAND_AT(insnobj, j) & (~1));
          rb_str_cat2(str, rb_id2name(entry->id));
        }
      case 'M': /* method cache */
        rb_str_cat2(str, "<ic>");
        break;
      case 'H': /* case/when condition cache */
        rb_str_cat2(str, "<ch>");
        break;
      default:{
        rb_bug("unknown operand type: %c", type);
      }
      }
      if(types[j+1]){
        rb_str_cat2(str, ", ");
      }
    }
  }
  return str;
}

VALUE insns_name_array(){
  VALUE ary = rb_ary_new();
  int i;
  for(i=0; i< sizeof(insn_name_info) / sizeof(insn_name_info[0]); i++){
    rb_ary_push(ary, rb_str_new2(insn_name_info[i]));
  }
  return ary;
}

void dump_disasm_list(struct iseq_link_element *link){
  int pos = 0;
  INSN *iobj;
  LABEL *lobj;
  VALUE str;

  printf("-- raw disasm--------\n");

  while(link){
    switch(link->type){
    case ISEQ_ELEMENT_INSN:{
      iobj = (INSN*)link;
      str = insn_data_to_s_detail(iobj);
      printf("%04d %-65s(%4d)\n",pos, StringValueCStr(str), insn_data_line_no(iobj));
      pos += insn_data_length(iobj);
      break;
    }
    case ISEQ_ELEMENT_LABEL:{
      lobj = (LABEL*)link;
      printf("<L%03d>\n", lobj->label_no);
      break;
    }
    case ISEQ_ELEMENT_NONE:{
      printf("[none]\n");
      break;
    }
    default:
      /* ignore */
      printf("%d\n", FIX2INT(link->type));
      rb_bug("dump_disasm_list error");
    }
    link = link->next;
  }
  printf("---------------------\n");
}
